"""
Load data from the Korean GEO-KOMPSAT-2A satellite.
    - Magnetometer data from SOSMAG instrument.
    - Particle data (electrons and protons) from KSEM particle instrument.

Data is downloaded from the ESA HAPI server, which requires authentication.
Users have to register with the ESA server 
    and replace their username and password in the code below.

Notes
-----
https://swe.ssa.esa.int/sosmag
https://swe.ssa.esa.int/hapi

Data types available:
    1. SOSMAG real time (datatype='1m', d3s_gk2a_sosmag_1m)
        Near-realtime Magnetic Field Data with 1-16Hz from SOSMAG
        on GEO-KOMPSAT-2A in geostationary orbit at 128.2E.
        'spase://SSA/NumericalData/D3S/d3s_gk2a_sosmag_1m'

    2. SOSMAG calibrated (default, datatype='', d3s_gk2a_sosmag_recalib)
        Recalibrated L2 Magnetic Field Data with 1-16Hz from SOSMAG
        on GEO-KOMPSAT-2A in geostationary orbit at 128.2E.
        'spase://SSA/NumericalData/D3S/d3s_gk2a_sosmag_recalib'

    3. Electrons real time:
        'spase://SSA/NumericalData/GEO-KOMPSAT-2A/kma_gk2a_ksem_pd_e_l1'
    4. Protons real time:
        'spase://SSA/NumericalData/GEO-KOMPSAT-2A/kma_gk2a_ksem_pd_p_l1'

"""

import logging
import requests
import json
import numpy as np
from pyspedas.tplot_tools import store_data, options
from pyspedas.tplot_tools import time_double, time_string, set_units, set_coords


# Global parameters for the functions in this file.
# Users must replace username and password with their ESA credentials.
esa_hapi_parameters = {
    "username": "spedas",  # Users must replace with their ESA username.
    "password": "acQ4pG6u9Bh26v2",  # User ESA password.
    "portal_url": "https://swe.ssa.esa.int/",
    "authenticate_url": "https://sso.ssa.esa.int/am/json/authenticate",
    "sso_cookiename": "iPlanetDirectoryPro",
    "print_errors": True,  # If true, prints error messages.
    "print_messages": False,  # If true, prints debugging/success messages.
}


def esa_get_auth_cookie():
    """
    Authenticates a user against OpenAM.

    Uses the esa_hapi_parameters dictionary for the parameters needed.
    Returns whether authentication was successful and the obtained cookie.
    If an error occurs, the exception is caught and the error printed.

    Parameters
    ----------
    None.

    Returns
    -------
    success: bool
        True, if authentication was successful
    auth_cookie: str
        The obtained authentication cookie
    """
    success = False
    auth_cookie = ""
    try:
        # Send a POST request to the authentication url
        response = requests.post(
            esa_hapi_parameters["authenticate_url"],
            headers={
                "Content-Type": "application/json",
                "X-OpenAM-Username": esa_hapi_parameters["username"],
                "X-OpenAM-Password": esa_hapi_parameters["password"],
            },
            data="{}",
        )
        # form the response, extract the auth cookie and return it
        token_dict = json.loads(response.content)
        auth_cookie = token_dict["tokenId"]
        success = True
    except Exception as exc:
        success = False
        auth_cookie = ""
        if esa_hapi_parameters["print_errors"]:
            logging.info(exc)
    finally:
        return success, auth_cookie


def esa_get_session(auth_cookie):
    """
    Establishes a session with the ESA HAPI server.

    Uses the authentication cookie obtained from esa_get_auth_cookie().
    Returns whether a session was established successfully and if so,
        the obtained session cookies.
    If an error occurs, the exception is caught and the error is printed.

    Parameters
    ----------
    auth_cookie: str
        The obtained authentication cookie from esa_get_auth_cookie()

    Returns
    -------
    success: bool
        True, if authentication was successful
    jsession_id: str
        The obtained session cookie
    xsrf_token: str
        The obtained xsrf token
    """
    success = False
    jsession_id = ""
    xsrf_token = ""
    try:
        # Try to access the HAPI/capabilities using the auth_cookie
        init_response = requests.get(
            esa_hapi_parameters["portal_url"] + "/hapi/capabilities",
            cookies={
                esa_hapi_parameters["sso_cookiename"]: auth_cookie,
            },
        )
        # Extract the session cookies from the very first response from HAPI
        cookie_jar = init_response.history[0].cookies
        jsession_id = cookie_jar.get("JSESSIONID")
        xsrf_token = cookie_jar.get("XSRF-TOKEN")
        # Extract the consent url we are being requested to send our consent to
        # (in case we didn't consent yet)
        consent_url = init_response.url
        content = init_response.content

        # If we consented already, we should have received HAPI/capabilities.
        if "/hapi/capabilities" not in consent_url:
            # If not, we need to give our consent in the next step.
            # Send the consent along with all cookies to the consent url.
            consent_response = requests.post(
                consent_url,
                cookies={
                    esa_hapi_parameters["sso_cookiename"]: auth_cookie,
                    "JSESSIONID": jsession_id,
                    "XSRF_TOKEN": xsrf_token,
                },
                data={
                    "decision": "Allow",
                    "save_consent": "on",
                },
            )
            content = consent_response.content
        # This will result in a redirect to the initial HAPI/capabilities.
        capabilities = json.loads(content)
        # The json output should be:
        #  2022/10/24: {'HAPI':'2.1.0','status':{'code':1200,'message':'OK'},'outputFormats':['csv','json']}
        #  previous: {'version':'2.1.0','status':{'code':1200,'message':'OK'},'outputFormats':['csv','json']}
        version = capabilities["HAPI"]
        status = capabilities["status"]
        # If the output is what we expect, return True and the session cookies
        if version != "" and status != {} and status["message"] == "OK":
            success = True
        else:
            success = False
            jsession_id = ""
            xsrf_token = ""
    except Exception as exc:
        if esa_hapi_parameters["print_errors"]:
            logging.info(exc)
        success = False
        jsession_id = ""
        xsrf_token = ""

    return success, jsession_id, xsrf_token


def esa_get_capabilities(jsession_id, xsrf_token):
    """
    Gets HAPI/capabilities from server.

    Uses the two cookies obtained using esa_get_session_cookies().
    If an error occurs, the exception is caught and the error printed.

    Parameters
    ----------
    jsession_id: str
        The obtained session cookie
    xsrf_token: str
        The obtained xsrf token

    Returns
    -------
    success: bool
        True, if function was successful
    capabilities: dict of str
        The obtained capabilities as a json dictionary of strings
    """
    success = False
    capabilities = {}
    try:
        # Send a GET request to the HAPI/capabilities.
        test_response = requests.get(
            esa_hapi_parameters["portal_url"] + "/hapi/capabilities",
            cookies={
                "JSESSIONID": jsession_id,
                "XSRF_TOKEN": xsrf_token,
            },
        )
        # Extract the capabilities from the response.
        capabilities = json.loads(test_response.content)
        version = capabilities["HAPI"]
        status = capabilities["status"]
        # If the capabilities are as expected, return True.
        if version != "" and status != {} and status["message"] == "OK":
            success = True
        else:
            success = False
            capabilities = {}
    except Exception as exc:
        if esa_hapi_parameters["print_errors"]:
            logging.info(exc)
        success = False
        capabilities = {}

    return success, capabilities


def esa_hapi_get_data(
    jsession_id,
    xsrf_token,  # HAPI session cookies.
    instrument="sosmag",  # Instrument name, either 'sosmag', 'electrons', or 'protons'
    datatype="",  # Data type for sosmag, either '' (recalibrated) or '1m' (real time)
    timemin="2022-01-31T01:00:00.000Z",  # Start time
    timemax="2022-01-31T01:10:00.000Z",  # End time
):
    """
    Gets either HAPI/data or HAPI/info from server.

    Uses the two cookies obtained from esa_get_session_cookies().
    If an error occurs, the exception is caught and the error printed.

    Parameters
    ----------
    jsession_id: str
        The obtained session cookie.
    xsrf_token: str
        The obtained xsrf token cookie.
    instrument: str
        The instrument name. Valid options: "sosmag", "electron", "protons".  Default: "sosmag"
    datatype: str
        Data type for sosmag, either '' (recalibrated) or '1m' (1 minute, real time).
    timemin: str
        Start time, for example: '2022-01-31T01:00:00.000Z'.
    timemax: str
        End time, for example: '2022-01-31T01:10:00.000Z'.

    Returns
    -------
    success: bool
        True, if function was successful.
    data: dict of str
        The obtained data as a json dictionary of strings.
    parameters: dict of str
        The obtained parameters for the data as a json dictionary of strings.
    description: str
        The obtained description of data.
    """
    success = False
    data = {}
    parameters = {}
    description = ""
    if instrument in ["electrons", "e"]:
        dataid = "spase://SSA/NumericalData/GEO-KOMPSAT-2A/kma_gk2a_ksem_pd_e_l1"
    elif instrument in ["protons", "p"]:
        dataid = "spase://SSA/NumericalData/GEO-KOMPSAT-2A/kma_gk2a_ksem_pd_p_l1"
    else:
        if datatype == "1m":
            dataid = "spase://SSA/NumericalData/D3S/d3s_gk2a_sosmag_1m"
        else:
            dataid = "spase://SSA/NumericalData/D3S/d3s_gk2a_sosmag_recalib"

    # This query string can also be used in a browser.
    hquery = (
        "data?id="
        + dataid
        + "&time.min="
        + timemin
        + "&time.max="
        + timemax
        + "&format=json"
    )
    fullurl = esa_hapi_parameters["portal_url"] + "/hapi/" + hquery
    if esa_hapi_parameters["print_messages"]:
        logging.info(fullurl)

    try:
        # send a GET request to the HAPI server
        test_response = requests.get(
            fullurl,
            cookies={
                "JSESSIONID": jsession_id,
                "XSRF_TOKEN": xsrf_token,
            },
        )
        # extract the content from the response
        dat = json.loads(test_response.content)
        data = dat["data"]
        parameters = dat["parameters"]
        description = dat["description"]
        status = dat["status"]

        # if the status are as expected, return True along with the data dict
        if data != "" and status != {} and status["message"] == "OK":
            success = True
        else:
            success = False
            data = {}
            parameters = {}
            description = ""
    except Exception as exc:
        if esa_hapi_parameters["print_errors"]:
            logging.info(exc)
        success = False
        data = {}
        parameters = {}
        description = ""

    return success, data, parameters, description


def sosmag_to_tplot(
    data,  # Data, array with 14 fields
    parameters,  # Description of data, array with 14 fields
    desc,  # Single string description of the data set
    datatype="",  # Either '' (recalibrated) or '1m' (real time)
    get_support_data=False,  # Download flags and other supporting data
    prefix="",  # To be added as prefix of tplot variable
    suffix="",  # To be added as suffix of tplot variable
):
    """
    Stores magnetometer data as tplot variables.

    Parameters
    ----------
    data: dict of str
        The obtained data as a json dictionary of strings.
    parameters: dict of str
        The obtained parameters for the data as a json dictionary of strings.
    desc: str
        The obtained description for this set of data.
    datatype: str
        Data type can be either '' (recalibrated) or '1m' (1 minute real time).
    get_support_data: bool
        If True, loads everything into tplot variables, including flags.
    prefix: str
        Prefix for tplot names. Default is ''.
    suffix: str
        Suffix for tplot names. Default is ''.

    Returns
    -------
    success: bool
        True, if function was successful.
    var_names: list of str
        Names of tplot variables created.
    """
    success = False
    var_names = []
    try:
        # Construct tplot variable names.
        if datatype == "":
            pre = prefix + "sosmag"
        else:
            pre = prefix + "sosmag_" + datatype

        # Get data
        d = np.array(data)  # data
        p = np.array(parameters)  # parameters

        if len(d.shape) != 2 or len(p.shape) != 1 or d.shape[1] != 14:
            logging.info("SOSMAG data has wrong shape. Abort.")
            return False

        # Time is the 0th field
        td = np.array(time_double(d[:, 0]))

        # Data version identifier
        # version, Data field: [1]
        if get_support_data:
            yd = np.array(d[:, 1], dtype=int)
            var_name1 = pre + "_version" + suffix
            pd = "Version"
            attr_dict = {"description": desc, "var_desc": pd}
            store_data(var_name1, data={"x": td, "y": yd}, attr_dict=attr_dict)
            options(var_name1, "legend_names", "version")
            options(var_name1, "ytitle", "version")
            options(var_name1, "ysubtitle", "")
            var_names.append(var_name1)

        # Magnetic field in GSE 'b_gse_x', 'b_gse_y', 'b_gse_z'
        # Data fields: [2, 3, 4]
        yd = np.array(d[:, 2:5], dtype=float)
        var_name2 = pre + "_b_gse" + suffix
        pnames = [p[2]["name"], p[3]["name"], p[4]["name"]]
        pd = "Magnetic Field B in GSE coordinates"
        attr_dict = {"description": desc, "var_desc": pd}
        store_data(var_name2, data={"x": td, "y": yd}, attr_dict=attr_dict)
        options(var_name2, "legend_names", pnames)
        options(var_name2, "ytitle", "b_gse")
        options(var_name2, "ysubtitle", "[nT]")
        set_coords(var_name2, "gse")
        set_units(var_name2, "nT")
        var_names.append(var_name2)

        # Magnetic field in HPEN 'b_hpen_x', 'b_hpen_y', 'b_hpen_z'
        # Data fields: [5, 6, 7]
        if get_support_data:
            yd = np.array(d[:, 5:8], dtype=float)
            var_name3 = pre + "_b_hpen" + suffix
            pnames = [p[5]["name"], p[6]["name"], p[7]["name"]]
            pd = "Magnetic Field B in HPEN coordinates"
            attr_dict = {"description": desc, "var_desc": pd}
            store_data(var_name3, data={"x": td, "y": yd}, attr_dict=attr_dict)
            options(var_name3, "legend_names", pnames)
            options(var_name3, "ytitle", "b_hpen")
            options(var_name3, "ysubtitle", "[nT]")
            set_coords(var_name3, "hpen")
            set_units(var_name3, "nT")
            var_names.append(var_name3)

        # Spacecraft Position in GSE 'position_x', 'position_y', 'position_z'
        # Data fields: [8, 9, 10]
        yd = np.array(d[:, 8:11], dtype=float)
        var_name4 = pre + "_position" + suffix
        pnames = [p[8]["name"], p[9]["name"], p[10]["name"]]
        pd = "Spacecraft Position in GSE"
        attr_dict = {"description": desc, "var_desc": pd}
        store_data(var_name4, data={"x": td, "y": yd}, attr_dict=attr_dict)
        options(var_name4, "legend_names", pnames)
        options(var_name4, "ytitle", "pos_gse")
        options(var_name4, "ysubtitle", "[km]")
        set_coords(var_name4,  "gse")
        set_units(var_name4, "km")
        var_names.append(var_name4)

        # Data flags
        # data_flags, Data field: [11]
        if get_support_data:
            yd = np.array(d[:, 11], dtype=int)
            var_name5 = pre + "_data_flags" + suffix
            pd = "Data Flags"
            attr_dict = {"description": desc, "var_desc": pd}
            store_data(var_name5, data={"x": td, "y": yd}, attr_dict=attr_dict)
            options(var_name5, "legend_names", p[11]["name"])
            options(var_name5, "ytitle", "data_flag")
            options(var_name5, "ysubtitle", "")
            var_names.append(var_name5)

        # Calibration status
        # final, Data_field: [12]
        if get_support_data:
            ydb = np.array(d[:, 12])  #  string
            yd1 = [0 if x == "0" or x == "false" or x == "False" else 1 for x in ydb]
            yd = np.array(yd1, dtype=bool)
            var_name6 = pre + "_final" + suffix
            pd = "Final"
            attr_dict = {"description": desc, "var_desc": pd}
            store_data(var_name6, data={"x": td, "y": yd}, attr_dict=attr_dict)
            options(var_name6, "legend_names", "final_flag")
            options(var_name6, "ytitle", "calib")
            options(var_name6, "ysubtitle", "")
            var_names.append(var_name6)

        # Current sampling frequency flag
        # frequency, Data field: [13]
        if get_support_data:
            yd = np.array(d[:, 13], dtype=int)
            var_name7 = pre + "_frequency_flag" + suffix
            pd = "Frequency flag"
            attr_dict = {"description": desc, "var_desc": pd}
            store_data(var_name7, data={"x": td, "y": yd}, attr_dict=attr_dict)
            options(var_name7, "legend_names", "freq_flag")
            options(var_name7, "ytitle", "freq_flag")
            options(var_name7, "ysubtitle", "")
            var_names.append(var_name7)

        # Print the variable names created
        logging.info("Tplot variables created!")
        logging.info(var_names)
        success = True
    except Exception as exc:
        if esa_hapi_parameters["print_errors"]:
            logging.info(exc)
        success = False

    return success, var_names


def particle_to_tplot(
    data,  # Data, array with 34 fields
    parameters,  # Description of data, array with 34 fields
    desc,  # Single string description of the data set
    ptype="p",  # Either 'p' or 'e'
    get_support_data=False,  # Download flags and other supporting data
    prefix="",  # To be added as prefix of tplot variable
    suffix="",  # To be added as suffix of tplot variable
):
    """
    Stores proton or electron flux data as tplot variables.

    Parameters
    ----------
    data: dict of str
        The obtained data as a json dictionary of strings.
    parameters: dict of str
        The obtained parameters for the data as a json dictionary of strings.
    desc: str
        The obtained description for this set of data.
    ptype: str
        Particle type can be either 'p' (protons) or 'e' (electrons).
    get_support_data: bool
        If True, loads everything into tplot variables, including flags.
    prefix: str
        Prefix for tplot names. Default is ''.
    suffix: str
        Suffix for tplot names. Default is ''.

    Returns
    -------
    success: bool
        True, if function was successful.
    var_names: list of str
        Names of tplot variables created.
    """
    success = False
    var_names = []
    try:
        # Construct tplot variable names.
        if len(prefix) != 0 and prefix[-1] != "_":
            prefix = prefix + "_"

        pre = prefix + "kompsat_" + ptype + "_"

        # Get data
        d = np.array(data)  # data
        p = np.array(parameters)  # parameters

        if len(d.shape) != 2 or len(p.shape) != 1 or d.shape[1] != 34:
            logging.info("Particle data has wrong shape. Abort.")
            return False

        # Time is the 0th field
        td = np.array(time_double(d[:, 0]))

        flux_bands = []
        if ptype == "e":
            # Electron flux 'e_flux'
            # Data fields: [1 - 11]
            # Ten energy bands for electrons
            flux_start = [
                "100",
                "150",
                "225",
                "325",
                "450",
                "700",
                "1350",
                "1800",
                "2600",
                "3800",
            ]
            for i in range(0, len(flux_start) - 1):
                flux_bands.append(flux_start[i] + "-" + flux_start[i + 1] + " keV")
            flux_bands.append("2000_3800 keV")  # Last band is 2000-3800 keV
        else:
            # Proton flux 'p_flux'
            # Data fields: [1 - 11]
            # Ten energy bands for protons
            flux_start = [
                "77",
                "119",
                "148",
                "229",
                "354",
                "548",
                "681",
                "1052",
                "2021",
                "3123",
                "6000",
            ]
            for i in range(0, len(flux_start) - 1):
                flux_bands.append(flux_start[i] + "-" + flux_start[i + 1] + " keV")

        for i in range(1, 11):
            yd = np.array(d[:, i], dtype=float)
            pname = p[i]["name"]  # e1, e2, ..., e10
            if ptype == "p":
                pname = pname.replace("e", "p")
            var_name0 = pre + pname + suffix
            pd = p[i]["description"]
            attr_dict = {"description": desc, "var_desc": pd}
            store_data(var_name0, data={"x": td, "y": yd}, attr_dict=attr_dict)
            options(var_name0, "legend_names", pname + " (" + flux_bands[i - 1] + ")")
            options(var_name0, "ytitle", pname)
            options(var_name0, "ysubtitle", "[cm^-2 sr^-1 s^-1 keV^-1]")
            set_coords(var_name0, "gse")
            set_units(var_name0,  "cm^-2 sr^-1 s^-1 keV^-1")
            var_names.append(var_name0)

        # Attenuator is opened ('0') or closed ('1')
        # att_flag, One data field: [11]
        if get_support_data:
            ydb = np.array(d[:, 11])
            yd1 = [0 if x == "0" or x == "false" or x == "False" else 1 for x in ydb]
            yd = np.array(yd1, dtype=bool)
            pname = p[11]["name"]
            var_name1 = pre + pname + suffix
            pd = "Attenuator flag"
            attr_dict = {"description": desc, "var_desc": pd}
            store_data(var_name1, data={"x": td, "y": yd}, attr_dict=attr_dict)
            options(var_name1, "legend_names", pname)
            options(var_name1, "ytitle", pname)
            options(var_name1, "ysubtitle", "")
            var_names.append(var_name1)

        # Indicates if detector is in normal status ('0') or disable status ('1')
        # det_flag1-8, Eight data fields: [12 - 19]
        if get_support_data:
            for i in range(12, 20):
                ydb = np.array(d[:, i])
                yd1 = [
                    0 if x == "0" or x == "false" or x == "False" else 1 for x in ydb
                ]
                yd = np.array(yd1, dtype=bool)
                pname = p[i]["name"]  # det_flag1, det_flag2, ..., det_flag8
                var_name2 = pre + pname + suffix
                pd = "Detector flag " + pname[-1]
                attr_dict = {"description": desc, "var_desc": pd}
                store_data(var_name2, data={"x": td, "y": yd}, attr_dict=attr_dict)
                options(var_name2, "legend_names", pname)
                options(var_name2, "ytitle", pname)
                options(var_name2, "ysubtitle", "")
                var_names.append(var_name2)

        # Quality flags, '0' means that the observation value is between min-value and max-value,
        # '1' means out of max-value, and '2' means out of min-value"
        # e1_qef-e10_qef, Ten data fields: [20 - 29]
        if get_support_data:
            for i in range(20, 30):
                yd = np.array(d[:, i], dtype=int)
                pname = p[i]["name"]  # e1_qef, e2_qef, ..., e10_qef
                var_name3 = pre + pname + suffix
                pd = "Observation value quality flag " + pname[-1]
                attr_dict = {"description": desc, "var_desc": pd}
                store_data(var_name3, data={"x": td, "y": yd}, attr_dict=attr_dict)
                options(var_name3, "legend_names", pname)
                options(var_name3, "ytitle", pname)
                options(var_name3, "ysubtitle", "")
                var_names.append(var_name3)

        # Number of acquisitions for 1-minute average data
        # integ_num, One data field: [30]
        if get_support_data:
            yd = np.array(d[:, 30], dtype=int)
            pname = p[30]["name"]  # integ_num
            var_name4 = pre + pname + suffix
            pd = "Number of acquisitions for 1-minute average data"
            attr_dict = {"description": desc, "var_desc": pd}
            store_data(var_name4, data={"x": td, "y": yd}, attr_dict=attr_dict)
            options(var_name4, "legend_names", pname)
            options(var_name4, "ytitle", pname)
            options(var_name4, "ysubtitle", "")
            var_names.append(var_name4)

        # Spacecraft Position in GSE 'position_x', 'position_y', 'position_z'
        # position_x, z, y, Three data fields: [31, 32, 33]
        yd = np.array(d[:, 31:34], dtype=float)
        var_name2 = pre + "pos_gse" + suffix
        pnames = [p[31]["name"], p[32]["name"], p[33]["name"]]
        pd = "Spacecraft Position in GSE"
        attr_dict = {"description": desc, "var_desc": pd}
        store_data(var_name2, data={"x": td, "y": yd}, attr_dict=attr_dict)
        options(var_name2, "legend_names", pnames)
        options(var_name2, "ytitle", "pos_gse")
        options(var_name2, "ysubtitle", "[km]")
        set_coords(var_name2,"gse")
        set_units(var_name2, "km")
        var_names.append(var_name2)

        # Print the variable names created
        logging.info("Variables created!")
        logging.info(var_names)
        success = True
    except Exception as exc:
        if esa_hapi_parameters["print_errors"]:
            logging.info(exc)
        success = False

    return success, var_names


def load(
    trange=[
        "2024-03-31T01:00:00.000Z",
        "2024-03-31T01:30:00.000Z",
    ],  # Default 30 minutes of data
    instrument="sosmag",  # Instrument name
    datatype="",  # Either '' (recalibrated) or '1m' (real time)
    get_support_data=False,  # Download flags and other supporting data
    prefix="",  # Prefix for tplot names
    suffix="",  # Suffix for tplot names
):
    """
    This function loads data from the ESA HAPI server.

    Parameters
    ----------
    trange: list of str
        Start and end times.
        Many time formats are supported, see function time_double().
    instrument: str
        Instrument name. Default is 'sosmag'.
        Valid values are: 'sosmag' or 'mag', 'electrons' or 'e', 'protons' or 'p'.
    datatype: str
        Either '' (recalibrated) or '1m' (real time).
        Valid only for sosmag data. Ignored for particle data.
    get_support_data: bool
        If True, loads everything into tplot variables, including flags.
    prefix: str
        Prefix for tplot names. Default is ''.
    suffix: str
        Suffix for tplot names. Default is ''.

    Returns
    -------
    var_names: list of str
        Names of tplot variables created.

    Examples
    --------
    >>> from pyspedas import tplot
    >>> from pyspedas import kompsat_load

    >>> # Plot L2 magnetometer data
    >>> var_names = kompsat_load(trange=["2024-03-31 02:00:00", "2024-03-31 03:00:00"], datatype="1m")
    >>> tplot(var_names)

    >>> # Plot electron data
    >>> var_names = kompsat_load(trange=["2024-03-31 02:00:00", "2024-03-31 03:00:00"], instrument="e")
    >>> tplot(var_names)

    >>> # Plot proton data
    >>> var_names = kompsat_load(trange=["2024-03-31 02:00:00", "2024-03-31 03:00:00"], instrument="p")
    >>> tplot(var_names)

    Notes
    -----
    Data links that can be tested on a browser:

    SOSMAG data:

    https://swe.ssa.esa.int/hapi/data?id=spase://SSA/NumericalData/D3S/d3s_gk2a_sosmag_recalib&time.min=2021-01-31T01:00:00.000Z&time.max=2021-01-31T01:01:00.000Z&format=json

    https://swe.ssa.esa.int/hapi/data?id=spase://SSA/NumericalData/D3S/d3s_gk2a_sosmag_1m&time.min=2021-01-31T01:00:00.000Z&time.max=2021-01-31T01:01:00.000Z&format=json

    Particle data:

    https://swe.ssa.esa.int/hapi/data?id=spase://SSA/NumericalData/GEO-KOMPSAT-2A/kma_gk2a_ksem_pd_p_l1&time.min=2024-03-31T01:00:00.000Z&time.max=2024-03-31T02:00:00.000Z&format=json

    https://swe.ssa.esa.int/hapi/data?id=spase://SSA/NumericalData/GEO-KOMPSAT-2A/kma_gk2a_ksem_pd_e_l1&time.min=2024-03-31T01:00:00.000Z&time.max=2024-03-31T02:00:00.000Z&format=json
    """
    var_names = []
    # Make sure that time is in the correct format for the ESA server.
    trange = time_string(time_double(trange), fmt="%Y-%m-%dT%H:%M:%S.000Z")
    if esa_hapi_parameters["print_messages"]:
        logging.info("trange", trange)

    # Authenticate with the HAPI server and receive a cookie.
    authenticated, auth_cookie = esa_get_auth_cookie()
    if authenticated:
        if esa_hapi_parameters["print_messages"]:
            logging.info("SOSMAG authentication successful.")
    else:
        logging.info(
            "SOSMAG authentication failed. \
        Please check username and password in the file sosmag:load.py."
        )
        return

    # Set a session with the server.
    capok = False
    if authenticated:
        sessionok, jsession_id, xsrf_token = esa_get_session(auth_cookie)
        if sessionok:
            # Session successfully established. Obtained session cookies.
            # Testing session cookies on hapi/capabilities.
            if esa_hapi_parameters["print_messages"]:
                logging.info("SOSMAG session established successfully.")
            capok, capabilities = esa_get_capabilities(jsession_id, xsrf_token)
            if capok:
                if esa_hapi_parameters["print_messages"]:
                    logging.info("SOSMAG HAPI server capabilities: ")
                    logging.info(capabilities)
            else:
                logging.info("Problem communicating with server. Aborting.")
                return
        else:
            logging.info("SOSMAG session could not be established. Aborting.")
            return

    # Get data.
    dataok = False
    if capok:
        dataok, data, parameters, desc = esa_hapi_get_data(
            jsession_id,
            xsrf_token,
            instrument=instrument,
            datatype=datatype,
            timemin=trange[0],
            timemax=trange[1],
        )
        if dataok:
            if instrument in ["electrons", "electron", "e", "protons", "proton", "p"]:
                ptype = instrument[0]  #'e' or 'p'
                tplotok, var_names = particle_to_tplot(
                    data,
                    parameters,
                    desc,
                    ptype=ptype,
                    get_support_data=get_support_data,
                    prefix=prefix,
                    suffix=suffix,
                )
            else:
                tplotok, var_names = sosmag_to_tplot(
                    data,
                    parameters,
                    desc,
                    datatype=datatype,
                    get_support_data=get_support_data,
                    prefix=prefix,
                    suffix=suffix,
                )
            if tplotok:
                if esa_hapi_parameters["print_messages"]:
                    logging.info("Data was loaded.")
                success = True
            else:
                logging.info("Could load data into pytplot. Aborting.")
                return
        else:
            logging.info("Could not get any data. Aborting.")
            return

    return var_names



from . import ace
from . import akebono
from . import barrel
from . import cluster
from . import cnofs
from . import csswe
from . import de2
from . import dscovr
from . import elfin
from . import equator_s
from . import erg
from . import fast
from . import geotail
from . import goes
from . import image
from . import kompsat
from . import kyoto
from . import lanl
from . import maven
from . import mica
from . import mms
from . import noaa
from . import omni
from . import poes
from . import polar
from . import psp
from . import rbsp
from . import secs
from . import soho
from . import solo
from . import st5
from . import stereo
from . import swarm
from . import themis
from . import twins
from . import ulysses
from . import wind


import os
import unittest
from pyspedas.tplot_tools import data_exists
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_load_mpa_data(self):
        out_vars = pyspedas.projects.lanl.mpa(time_clip=True)
        self.assertTrue(data_exists("dens_lop"))
        self.assertTrue(data_exists("vel_lop"))
        self.assertTrue(data_exists("temp_lop"))

    def test_load_spa_data(self):
        out_vars = pyspedas.projects.lanl.spa(time_clip=True)
        self.assertTrue(data_exists("spa_p_temp"))
        self.assertTrue(data_exists("spa_e_temp"))
        self.assertTrue(data_exists("spa_a_flux"))
        self.assertTrue(data_exists("spa_e_flx"))
        self.assertTrue(data_exists("spa_p_flx"))

    def test_load_notplot(self):
        out_vars = pyspedas.projects.lanl.spa(notplot=True)
        self.assertTrue("spa_p_temp" in out_vars)

    def test_downloadonly(self):
        files = pyspedas.projects.lanl.spa(
            downloadonly=True, trange=["2004-10-31", "2004-11-01"]
        )
        self.assertTrue(os.path.exists(files[0]))


if __name__ == "__main__":
    unittest.main()



import numpy as np


def slice2d_checkbins(dist1, dist2):
    """
    Checks if two particle distribution structures have identical
    energy, phi, theta, and mass values.

    Input
    ------
        dist1: 3D particle data structure
            First particle data structure

        dist2: 3D particle data structure
            Second particle data structure

    Output
    -------
        Returns 1 if all fields match and 0 otherwise
    """

    if dist2 is None:
        return 1

    if not np.array_equal(dist1['phi'], dist2['phi']) or \
       not np.array_equal(dist1['theta'], dist2['theta']) or \
       not np.array_equal(dist1['energy'], dist2['energy']) or \
       dist1['mass'] != dist2['mass']:
        return 0

    return 1



from .datasets import datasets
from .efw import efw
from .emfisis import emfisis
from .hope import hope
from .mageis import mageis
from .magephem import magephem
from .rbspice import rbspice
from .rbspice_lib.rbsp_rbspice_omni import rbsp_rbspice_omni
from .rbspice_lib.rbsp_rbspice_pad import rbsp_rbspice_pad
from .rept import rept
from .rps import rps



# Copyright 2020 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import pandas as pd
import numpy as np
import datetime
import logging
from pyspedas.tplot_tools import del_data, tplot_rename, get_y_range, replace_metadata
import pyspedas
import xarray as xr
import copy
import warnings

tplot_num = 1


def store_data(name, data=None, delete=False, newname=None, attr_dict={}):
    
    """
    Create a "Tplot Variable" (similar to the IDL SPEDAS concept) based on the inputs, and
    stores this data in memory.  Tplot Variables store all of the information
    needed to generate a plot.  
    
    Parameters
    ----------
        name : str 
            Name of the tplot variable that will be created
        data : dict or list[str]
            A python dictionary object for creating a single variable, or a list of base variables to combine them into a 'pseudovariable'
            
            'x' should be a 1-dimensional array that represents the data's x axis.  If x is a numeric type, it is interpreted
            as seconds since the Unix epoch.  x can also be passed as Pandas Series object, datetime.datetime, numpy.datetime64, or strings.
            represented in seconds since epoch (January 1st 1970)
            
            'y' should be the data values. This can be 2 dimensions if multiple lines or a spectrogram are desired.
            
            'v' is optional, and is only used for spectrogram plots.  This will be a list of bins to be used.  If this
            is provided, then 'y' should have dimensions of x by z.

            'v1/v2/v3/etc' are also optional, and are only used for to spectrogram plots.  These will act as the coordinates
            for 'y' if 'y' has numerous dimensions.  By default, 'v2' is plotted in spectrogram plots.

        delete : bool, optional
            If True, deletes the tplot variable matching the "name" parameter
            Default: False
        newname: str
            If set, renames TVar to new name
            Default: False
        attr_dict: dict
            A dictionary object of attributes (these do not affect routines in pyspedas, this is merely to keep metadata alongside the file)
            Default: {} (empty dictionary)
        
    .. note::
        If you want to combine multiple tplot variables into one, simply supply the list of tplot variables to the
        "data" parameter.  This will cause the data to overlay when plotted.
        
    Returns
    -------
        bool
            True if successful, False otherwise
        
    Examples
    --------
        >>> # Store a single line
        >>> import pyspedas
        >>> x_data = [1,2,3,4,5]
        >>> y_data = [1,2,3,4,5]
        >>> pyspedas.store_data("Variable1", data={'x':x_data, 'y':y_data})
    
        >>> # Store two lines
        >>> x_data = [1,2,3,4,5]
        >>> y_data = [[1,5],[2,4],[3,3],[4,2],[5,1]]
        >>> pyspedas.store_data("Variable2", data={'x':x_data, 'y':y_data})
        
        >>> # Store a spectrogram
        >>> x_data = [1,2,3]
        >>> y_data = [ [1,2,3] , [4,5,6], [7,8,9] ]
        >>> v_data = [1,2,3]
        >>> pyspedas.store_data("Variable3", data={'x':x_data, 'y':y_data, 'v':v_data})
        
        >>> # Combine two different line plots
        >>> pyspedas.store_data("Variable1and2", data=['Variable1', 'Variable2'])
        
        >>> #Rename TVar
        >>> pyspedas.store_data('a', data={'x':[0,4,8,12,16], 'y':[1,2,3,4,5]})
        >>> pyspedas.store_data('a',newname='f')

    """
    
    # global tplot_num
    create_time = datetime.datetime.now()
    # If delete is specified, we are just deleting the variable
    if delete is True:
        del_data(name)
        return False

    if data is None and newname is None and attr_dict is None:
        logging.error('store_data: data array, newname, and attr_dict all unspecified, nothing to do.')
        return False

    if data is None and newname is None and attr_dict is not None:
        replace_metadata(name,attr_dict)
        return True

    # If newname is specified, we are just renaming the variable
    if newname is not None:
        tplot_rename(name, newname)
        return True

    # if isinstance(data, str):
    #     pyspedas.tplot_tools.data_quants[name] = {'name': name, 'data': data}
    #     return True
    if isinstance(data, str):
        data = data.split(' ')

    # If the data is a list instead of a dictionary, user is looking to overplot
    if isinstance(data, list):
        base_data = _get_base_tplot_vars(name,data)
        if len(base_data) == 0:
            logging.warning("store_data: None of the base variables exist to construct pseudovariable %s",name)
            return False
        # Copying the first variable to use all of its plot options
        # However, we probably want each overplot to retain its original plot option
        pyspedas.tplot_tools.data_quants[name] = copy.deepcopy(pyspedas.tplot_tools.data_quants[base_data[0]])
        pyspedas.tplot_tools.data_quants[name].attrs = copy.deepcopy(pyspedas.tplot_tools.data_quants[base_data[0]].attrs)
        pyspedas.tplot_tools.data_quants[name].name = name
        pyspedas.tplot_tools.data_quants[name].attrs['plot_options']['overplots'] = base_data[1:]
        pyspedas.tplot_tools.data_quants[name].attrs['plot_options']['overplots_mpl'] = base_data
        # These sets of options should default to the sub-variables' options, not simply
        # copied from the first variable in the list.   These options can be still be set
        # on the pseudovariable, and they will override the sub-variable options.
        pyspedas.tplot_tools.data_quants[name].attrs['plot_options']['yaxis_opt'] = {}
        pyspedas.tplot_tools.data_quants[name].attrs['plot_options']['zaxis_opt'] = {}
        pyspedas.tplot_tools.data_quants[name].attrs['plot_options']['line_opt'] = {}
        pyspedas.tplot_tools.data_quants[name].attrs['plot_options']['extras'] = {}
        return True

    # if the data table doesn't contain an 'x', assume this is a non-record varying variable
    if 'x' not in data.keys():
        values = np.array(data.pop('y'))
        pyspedas.tplot_tools.data_quants[name] = {'data': values}
        pyspedas.tplot_tools.data_quants[name]['name'] = name
        return True

    times = data.pop('x')

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        values = np.array(data.pop('y'))

    if 'dy' in data.keys():
        err_values = np.array(data.pop('dy'))

        if len(err_values) != len(times):
            logging.warning('store_data: Warning: %s: length of error values (%d) does not match length of time values (%d)',name,len(err_values),
                            len(times))
    else:
        err_values = None

    # Convert input time representation to np.datetime64 objects, if needed
    if isinstance(times, pd.Series):
        datetimes = times.to_numpy()  # if it is pandas series, convert to numpy array
    elif isinstance(times[0],datetime.datetime):
        # Timezone-naive datetime, use as-is, but we might have to convert the container to a numpy array
        if isinstance(times,np.ndarray):
            datetimes = times
        else:
            datetimes = np.array(times)
    elif isinstance(times[0],np.datetime64):
        # np.datetime64, use as-is, but we might have to convert the container to a numpy array
        if isinstance(times,np.ndarray):
            datetimes = times
        else:
            datetimes = np.array(times)
        # We want the np.datetime64 resolution to be ns.  If it already is, do nothing, otherwise,
        # convert to ns.  In the future, we might support storing times in any resolution,
        # and dealing with the conversion in get_data or in client code.
        dtype = datetimes.dtype
        if dtype.name != 'datetime64[ns]':
            datetimes = datetimes.astype('datetime64[ns]')
    elif isinstance(times[0],(int,np.integer,float,np.float64)):
        # Assume seconds since Unix epoch, convert to np.datetime64 with nanosecond precision
        # Make sure we have a numpy array
        if not isinstance(times,np.ndarray):
            times=np.array(times)
        # Replace any NaN or inf values with 0
        cond = np.logical_not(np.isfinite(times))
        times[cond] = 0
        datetimes = np.array(times*1e09,dtype='datetime64[ns]')
    elif isinstance(times[0],str):
        # Interpret strings as timestamps, convert to np.datetime64 with nanosecond precision
        datetimes = np.array(times,dtype='datetime64[ns]')
    else:
        # Hope it's convertable to a numpy array!  This case will get hit for an xarray DataArray.
        datetimes = np.array(times)

    times = datetimes

    # At this point, times should be a numpy array of datetime or np.datetime64 objects

    if len(values.shape) == 0:
        # This can happen for Cluster variables with only a single sample, as they can
        # be incorrectly marked as NRV and lose their leading (time) dimension.
        logging.warning("store_data: Data array for %s appears to be a zero-dimensional array; converting to 1-D array.",name)
        if len(times) == 1:
            logging.warning("store_data: This is possibly due to the leading array dimension being lost in a scalar variable with a single timestamp.")
        values = np.array([values])

    if len(values) == 0:
        logging.warning('store_data: %s has empty y component, cannot create variable',name)
        return False

    if len(times) != len(values):
        # This happens for a few MMS and other data sets. Rather than quitting immediately, go ahead and create
        # the variable, but give an informational message about the mismatch.  The fix would probably be for the
        # data provider to mark the variable as non-record-variant, and avoid giving it a DEPEND_0 or DEPEND_TIME
        # attribute.
        logging.info("store_data: %s: lengths of x (%d) and y (%d) do not match! Mislabeled NRV variable?",name,len(times),len(values))

    if not isinstance(times,np.ndarray):
        logging.warning("store_data: times was not converted to a numpy array. This should not happen.")
        times = np.array(times)

    # assumes monotonically increasing time series
    if isinstance(times[0], datetime.datetime):
        # This may be dead code now?
        trange = [times[0].replace(tzinfo=datetime.timezone.utc).timestamp(),
                  times[-1].replace(tzinfo=datetime.timezone.utc).timestamp()]
    elif isinstance(times[0], np.datetime64):
        trange = np.float64([times[0], times[-1]]) / 1e9
    else:
        trange = [times[0], times[-1]]

    # Special case if y is 1-dimensional and 'v' or 'v1' is present
    # This can happen if split_data is called on a vector-valued variable that has a DEPEND_1.
    # We can't use v as a coordinate, or we'll get a ValueError creating the xarray object,
    # so we'll save its value here, then after the xarray object is created, stash it in a different
    # attribute.  Then join_vec can find it and restore the depend_1 array from split-out components.
    extra_v_values = None
    if len(values.shape) == 1:
        if 'v' in data.keys():
            extra_v_values = np.array(data.pop('v'))
        elif 'v1' in data.keys():
            extra_v_values = np.array(data.pop('v1'))

    # Figure out the 'v' data
    # This seems to be conflating specplot bins with general DEPEND_N attributes.
    # Maybe only do this stuff if it's marked as a spectrum?  But what if it's from
    # a NetCDF rather than a CDF?
    spec_bins_exist = False
    if 'v' in data or 'v1' in data or 'v2' in data or 'v3' in data:
        # Generally the data is 1D, but occasionally
        # the bins will vary in time.
        spec_bins_exist = True
        if 'v' in data:
            spec_bins = data['v']
            spec_bins_dimension = 'v'
        elif ("v1" in data) and ("v2" in data) and ("v3" in data):
            spec_bins = data['v2']
            spec_bins_dimension = 'v2'
        elif ("v1" in data) and ("v2" in data):
            spec_bins = data['v2']
            spec_bins_dimension = 'v2'
        else:
            # At least one vn is missing.
            logging.warning("At least one Vn tag is missing, cannot create spec_bins from variable %s.", name)
            spec_bins_exist = False

        if spec_bins_exist and type(spec_bins) is not pd.DataFrame:
            try:
                spec_bins = pd.DataFrame(spec_bins)
            except:
                if spec_bins_dimension=='v':
                    spec_bins = np.arange(1, len(values[0])+1)
                elif spec_bins_dimension=="v2":
                    spec_bins = np.arange(1, len(values[0][0]) + 1)
                elif spec_bins_dimension=="v3":
                    spec_bins = np.arange(1, len(values[0][0][0]) + 1)
                spec_bins = pd.DataFrame(spec_bins)


        if spec_bins_exist and len(spec_bins.columns) != 1:
            # The spec_bins are time varying
            # Or maybe they're just DEPEND_N and nothing to do with spectra?
            spec_bins_time_varying = True
            if len(spec_bins) != len(times):
                # Maybe it's not a spectrum at all?
                # Cluster pressure tensor variablea havw a DEPEND_1 that's 2-D, 1x3 [['x','y','z']]
                logging.error("store_data: Length of spec_bins (%d) and times (%d) do not match for variable %s.",len(spec_bins),len(times),name)
                spec_bins = None
                spec_bins_exist = False
        elif spec_bins_exist:
            spec_bins = spec_bins.transpose()
            spec_bins_time_varying = False
    else:
        spec_bins = None
        # Provide another dimension if values are more than 1 dimension
        if len(values.shape) == 2:
            data['v'] = None
        if len(values.shape) > 2:
            data['v1'] = None
            data['v2'] = None
        if len(values.shape) > 3:
            data['v3'] = None

    # Set up xarray dimension and coordinates
    coordinate_list = sorted(list(data.keys()))
    dimension_list = [d + '_dim' for d in coordinate_list]

    if len(coordinate_list) < len(values.shape)-1:
        logging.warning("store_data: Data array for variable %s has %d dimensions, but only %d v_n keys plus time. Adding empty v_n keys.", name, len(values.shape), len(coordinate_list))
        if len(values.shape) == 2:
            data['v'] = None
        elif len(values.shape) == 3:
            if 'v' in data:
                vdat = data.pop('v')
                data['v1'] = vdat
            elif 'v1' in data:
                pass
            if 'v1' not in data:
                data['v1'] = None
            if 'v2' not in data:
                data['v2'] = None
        elif len(values.shape) == 4:
            # ERG LEPI 3dflux quality flags have this issue
            if 'v' in data:
                vdat = data.pop('v')
                data['v1'] = vdat
            elif 'v1' in data:
                pass
            if 'v1' not in data:
                data['v1'] = None
            if 'v2' not in data:
                data['v2'] = None
            if 'v3' not in data:
                data['v3'] = None

        coordinate_list = sorted(list(data.keys()))
        dimension_list = [d + '_dim' for d in coordinate_list]
        # Don't try to use these dimensions as coordinates
        spec_bins_exist = False
        spec_bins = None

    temp = None
    # Ignore warnings about cdflib non-nanosecond precision timestamps for now
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore",message="^.*non-nanosecond precision.*$")
        try:
            temp = xr.DataArray(values, dims=['time']+dimension_list,
                                coords={'time': ('time', times)})
        except ValueError as err:
            logging.warning("store_data: ValueError trying to set xarray coordinates for variable %s: %s", name, str(err))
            spec_bins_exist = False
            spec_bins = None
            if len(times) == 1:
                logging.warning("store_data: This is possibly due to the leading data dimension being lost in an array-valued or vector-valued variable with a single timestamp.")
            # If data is 1-dimensional, ignore any DEPEND_N supplied
            elif (len(values.shape) == 1) and len(dimension_list) > 0:
                logging.warning("store_data: variable %s is 1-dimensional, but has additional keys defined: %s.  Dropping redundant coordinate(s).",name, dimension_list)
                temp = xr.DataArray(values, dims=['time'], coords={'time': ('time', times)})
                coordinate_list=[]
                dimension_list=[]
            else:
                logging.warning("Giving up on this variable.")
                return

    if temp is None:
        # This can happen with mismatched times/data values, and no valid DEPEND_N.
        # For example, POLAR MFE data, variable MF_Num
        logging.warning("store_data: Unable to create xarray object for variable %s, giving up.", name)
        return

    if spec_bins_exist:
        try:
            if spec_bins_time_varying:
                temp.coords['spec_bins'] = (('time', spec_bins_dimension+'_dim'), spec_bins.values)
            else:
                temp.coords['spec_bins'] = (spec_bins_dimension+'_dim', np.squeeze(spec_bins.values))
        except ValueError as err:
            logging.warning('store_data: conflicting size for at least one dimension for variable %s', name)
            logging.warning('store_data: ValueError exception text: %s',str(err))

    for d in coordinate_list:
        if data[d] is None:
            continue
        try:
            d_dimension = pd.DataFrame(data[d])
            if len(d_dimension.columns) != 1:
                if len(d_dimension) != len(times):
                    logging.warning("store_data: Length of %s (%d) and time (%d) do not match.  Cannot create coordinate for %s.",d,len(d_dimension),len(times),name)
                    continue
                temp.coords[d] = (('time', d+'_dim'), d_dimension.values)
            else:
                d_dimension = d_dimension.transpose()
                squeezed_array = np.squeeze(d_dimension.values)# np.squeeze() does something funny here if this dimension has length 1, causing a ValueError exception
                if d_dimension.size == 1:
                    logging.warning("store_data: Dimension %s of variable %s has length 1",d,name)
                    temp.coords[d] = (d+'_dim', d_dimension.values[0])
                else:
                    temp.coords[d] = (d+'_dim', squeezed_array)
        except ValueError as err:
            logging.warning("store_data: Could not create coordinate %s_dim for variable %s",d, name)
            logging.warning("store_data: ValueError exception text: %s", str(err))

    # Set up Attributes Dictionaries
    xaxis_opt = dict(axis_label='')
    yaxis_opt = dict(axis_label=name) if (spec_bins is None) else dict(axis_label='')
    zaxis_opt = dict(axis_label='Z-Axis') if (spec_bins is None) else dict(axis_label=name)
    xaxis_opt['crosshair'] = 'X'
    yaxis_opt['crosshair'] = 'Y'
    zaxis_opt['crosshair'] = 'Z'
    xaxis_opt['x_axis_type'] = 'linear'
    yaxis_opt['y_axis_type'] = 'linear'
    zaxis_opt['z_axis_type'] = 'linear'
    line_opt = {}
    time_bar = []
    extras = dict(panel_size=1, border=True)
    links = {}

    # Add dicts to the xarray attrs
    temp.name = name
    temp.attrs = copy.deepcopy(attr_dict)
    if extra_v_values is not None:
        temp.attrs['extra_v_values'] = extra_v_values

    if 'plot_options' not in temp.attrs.keys():
        temp.attrs['plot_options'] = {}
        temp.attrs['plot_options']['xaxis_opt'] = xaxis_opt
        temp.attrs['plot_options']['yaxis_opt'] = yaxis_opt
        temp.attrs['plot_options']['zaxis_opt'] = zaxis_opt
        temp.attrs['plot_options']['line_opt'] = line_opt
        temp.attrs['plot_options']['trange'] = trange
        temp.attrs['plot_options']['time_bar'] = time_bar
        temp.attrs['plot_options']['extras'] = extras
        temp.attrs['plot_options']['create_time'] = create_time
        temp.attrs['plot_options']['links'] = links
        #temp.attrs['plot_options']['spec_bins_ascending'] = _check_spec_bins_ordering(times, spec_bins)
        temp.attrs['plot_options']['overplots'] = []
        temp.attrs['plot_options']['overplots_mpl'] = []
        temp.attrs['plot_options']['interactive_xaxis_opt'] = {}
        temp.attrs['plot_options']['interactive_yaxis_opt'] = {}
        temp.attrs['plot_options']['error'] = err_values

    pyspedas.tplot_tools.data_quants[name] = temp

    pyspedas.tplot_tools.data_quants[name].attrs['plot_options']['yaxis_opt']['y_range'] = get_y_range(temp)

    return True


def _get_base_tplot_vars(name,data):
    base_vars = []
    if not isinstance(data, list):
        data = [data]
    for var in data:
        if var not in pyspedas.tplot_tools.data_quants:
            logging.warning('store_data: Pseudovariable %s component %s not found, skipping', name, var)
        elif isinstance(pyspedas.tplot_tools.data_quants[var].data, list):
            base_vars += _get_base_tplot_vars(name,pyspedas.tplot_tools.data_quants[var].data)
        else:
            base_vars += [var]
    return base_vars


def _check_spec_bins_ordering(times, spec_bins):
    """
    This is a private function, this is run during
    object creation to check if spec_bins are ascending or descending
    """
    if spec_bins is None:
        return
    if len(spec_bins) == len(times):
        break_top_loop = False
        for index, row in spec_bins.iterrows():
            if row.isnull().values.all():
                continue
            else:
                for i in row.index:
                    if np.isfinite(row[i]) and np.isfinite(row[i + 1]):
                        ascending = row[i] < row[i + 1]
                        break_top_loop = True
                        break
                    else:
                        continue
                if break_top_loop:
                    break
    else:
        ascending = spec_bins[0].iloc[0] < spec_bins[1].iloc[0]
    return ascending


def store(name, data=None, delete=False, newname=None, metadata={}):
    """
    Create tplot variables. This is a wrapper for store_data, with the only apparent
    difference being that 'attr_dict' in store_data is replaced with 'metadata' in store().
    This wrapper will likely be removed in a future release.
    Parameters:
        name : str
            Name of the tplot variable that will be created
        data : dict
            A python dictionary object.

            'x' should be a 1-dimensional array that represents the data's x axis.  Typically this data is time,
            represented in seconds since epoch (January 1st 1970)

            'y' should be the data values. This can be 2 dimensions if multiple lines or a spectrogram are desired.

            'v' is optional, and is only used for spectrogram plots.  This will be a list of bins to be used.  If this
            is provided, then 'y' should have dimensions of x by z.

            'v1/v2/v3/etc' are also optional, and are only used for to spectrogram plots.  These will act as the coordinates
            for 'y' if 'y' has numerous dimensions.  By default, 'v2' is plotted in spectrogram plots.

            'x' and 'y' can be any data format that can be read in by the pandas module.  Python lists, numpy arrays,
            or any pandas data type will all work.
        delete : bool, optional
            Deletes the tplot variable matching the "name" parameter
        newname: str
            Renames TVar to new name
        metadata: dict
            A dictionary object of attributes (these do not affect routines in pyspedas, this is merely to keep metadata alongside the file)

    .. note::
        If you want to combine multiple tplot variables into one, simply supply the list of tplot variables to the
        "data" parameter.  This will cause the data to overlay when plotted.

    Returns:
        None

    Examples:
        >>> # Store a single line
        >>> import pyspedas
        >>> x_data = [1,2,3,4,5]
        >>> y_data = [1,2,3,4,5]
        >>> pyspedas.store("Variable1", data={'x':x_data, 'y':y_data})

        >>> # Store a two lines
        >>> x_data = [1,2,3,4,5]
        >>> y_data = [[1,5],[2,4],[3,3],[4,2],[5,1]]
        >>> pyspedas.store("Variable2", data={'x':x_data, 'y':y_data})

        >>> # Store a spectrogram
        >>> x_data = [1,2,3]
        >>> y_data = [ [1,2,3] , [4,5,6], [7,8,9] ]
        >>> v_data = [1,2,3]
        >>> pyspedas.store("Variable3", data={'x':x_data, 'y':y_data, 'v':v_data})

        >>> # Combine two different line plots
        >>> pyspedas.store("Variable1and2", data=['Variable1', 'Variable2'])

        >>> #Rename TVar
        >>> pyspedas.store('a', data={'x':[0,4,8,12,16], 'y':[1,2,3,4,5]})
        >>> pyspedas.store('a',newname='f')
    """
    return store_data(name, data=data, delete=delete, newname=newname, attr_dict=metadata)




import os
import copy
import calendar
import logging
import numpy as np
import xarray as xr
import pyspedas
from pyspedas.tplot_tools import tplot, store_data
from netCDF4 import Dataset, num2date


def change_time_to_unix_time(time_var):
    """
    Convert the variable to seconds since epoch.
    """
    # Capitalization of variable attributes may vary...
    if hasattr(time_var, "units"):
        units = time_var.units
    elif hasattr(time_var, "Units"):
        units = time_var.Units
    elif hasattr(time_var, "UNITS"):
        units = time_var.UNITS
    # ICON uses nonstandard units strings
    if units == "ms":
        units = "milliseconds since 1970-01-01 00:00:00"
    dates = num2date(time_var[:], units=units)
    unix_times = list()
    for date in dates:
        unix_time = calendar.timegm(date.timetuple())
        unix_times.append(unix_time)
    return unix_times


def netcdf_to_tplot(
    filenames, time="", prefix="", suffix="", plot=False, merge=False, strict_time=True
):
    """
    Create tplot variables from netCDF files.

    Parameters
    ----------
    filenames : str or list of str
        The file names and full paths of netCDF files.
    time : str, optional
        This is not used anymore. Remains here for backward compatibility.
        Currently, the name of the time variable is found in the netCDF variables themselves.
    prefix : str, optional
        The tplot variable names will be given this prefix.
        By default, no prefix is added.
    suffix : str, optional
        The tplot variable names will be given this suffix.
        By default, no suffix is added.
    plot : bool, optional
        If True, the data is plotted immediately after being generated. All tplot
        variables generated from this function will be on the same plot.
        By default, a plot is not created.
    merge : bool, optional
        If True, then data from 'filenames' will be merged into existing tplot variables.
        If False (default), then data from 'filenames' will overwrite existing tplot variables.
        Data in 'filenames' will always be merged/combined by themselves.
    strict_time : bool, optional
        If True (default), variables will be loaded into tplot variables only if
        their data length matches the time length.
        If False, all variables will be loaded. This is useful because some
        variables may contain general information, like satellite longitude.

    Returns
    -------
    stored_variables : list of str
        List of tplot variables created.

    Examples
    --------
    Create tplot variables from a GOES netCDF file:

    >>> import pyspedas
    >>> file = "/Users/user_name/goes_files/g15_epead_a16ew_1m_20171201_20171231.nc"
    >>> pyspedas.netcdf_to_tplot(file, prefix='mvn_')

    Add a prefix, and plot immediately:

    >>> import pyspedas
    >>> file = "/Users/user_name/goes_files/g15_epead_a16ew_1m_20171201_20171231.nc"
    >>> pyspedas.netcdf_to_tplot(file, prefix='goes_prefix_', plot=True)
    """

    stored_variables = []

    if prefix is None:
        prefix = ''
    if suffix is None:
        suffix = ''

    if isinstance(filenames, str):
        filenames = [filenames]
    elif isinstance(filenames, list):
        pass
    else:
        logging.error("Invalid filenames input. Must be string or list of strings.")
        return stored_variables

    filenames = sorted(list(set(filenames)))
    for filename in filenames:

        # Read file
        if os.path.isfile(filename):
            vfile = Dataset(filename)
        else:
            logging.error("Cannot find file: " + filename)
            continue

        # Create a dictionary that contains variables and their attributes.
        vars_and_atts = {}
        for name, variable in vfile.variables.items():
            vars_and_atts[name] = {}
            for attrname in variable.ncattrs():
                vars_and_atts[name][attrname] = getattr(variable, attrname)

        # Fill in missing values for each variable with np.nan (if values are not already nan)
        # and save the masked variables to a new dictionary.
        masked_vars = {}  # Dictionary containing properly masked variables
        for var in vars_and_atts.keys():
            reg_var = vfile.variables[var]
            # Check for some attributes that might be used to flag fill values
            atts_dict = vars_and_atts[var]
            fillval_atts_lc = ["fillval", "_fillval", "_fillvalue", "fillvalue", "missing_data"]
            var_fill_value = None
            for key in atts_dict.keys():
                if key.lower() in fillval_atts_lc:
                    # If multiple matching keys are found, the one that appears latest in the above list
                    # will take precedence
                    var_fill_value = atts_dict[key]

            # If var_fill_value is None, or already NaN, there's nothing to do here.
            # Integer arrays can't be NaN-filled, so if var_fill_value is any kind of integer, skip those too.
            # Some missions have strings defined as fill values.  (ICON)
            if var_fill_value is not None and not isinstance(var_fill_value, np.integer) and not isinstance(var_fill_value, str) and not np.isnan(var_fill_value):
                # We want to force missing values to be nan so that plots don't look strange
                var_mask = np.ma.masked_where(
                    reg_var == np.float32(var_fill_value), reg_var
                )
                var_filled = np.ma.filled(var_mask, np.nan)
                masked_vars[var] = var_filled
            else:
                var_filled = reg_var
                masked_vars[var] = var_filled

        # A dictionary with the time variables in this file.
        times_dict = {}

        # Store each netcdf variable as a tplot variable.
        for i, var in enumerate(vfile.variables):

            # Make sure that the variables are time-based, otherwise don't store them as tplot variables.
            if len(vfile[var].dimensions) > 0 and len(vfile[var].dimensions[0]) > 0:

                # Find the time dependence of the current variable.
                this_time = vfile[var].dimensions[0]
                if this_time not in vars_and_atts.keys():
                    # For GOES satelites, sometimes we get 'record' as time dependance.
                    # In that case, we can try 'time' and 'time_tag' as alternatives.
                    if "time" in vars_and_atts.keys():
                        this_time = "time"
                    elif "time_tag" in vars_and_atts.keys():
                        this_time = "time_tag"

                if this_time not in vars_and_atts.keys():
                    # If this_time does not exist, we can't save this as tplot variable.
                    continue
                elif this_time == var:
                    # If this_time has the same name as the current variable, do not save it.
                    continue

                # Find the time values (as unix times).
                if this_time in times_dict:
                    unix_times = times_dict[this_time]
                else:
                    try:
                        time_var = vfile[this_time]
                        unix_times = change_time_to_unix_time(time_var)
                        times_dict[this_time] = unix_times
                    except Exception as e:
                        # In this case, we could not handle the time, print an error
                        logging.error(
                            "Could not process time variable '"
                            + this_time
                            + "' for the netcdf variable: '"
                            + var
                            + "'"
                        )
                        logging.error("Exception details: " + str(e))
                        continue

                if var not in masked_vars:
                    # We don't have any values for this variable, skip it.
                    continue
                this_masked_var = masked_vars[var]
                if len(this_masked_var.shape) < 1:
                    # Values are empty, skip it.
                    continue
                if len(unix_times) != this_masked_var.shape[0] and strict_time:
                    # If strict_time is true, reject all variables that do not have
                    # same length for data and time. These can be inclination and other information
                    # saved as netcdf variables.
                    # If strict_time is false, pyspedas.store_data will complain about this
                    # "lengths of x and y do not match", but it will create the tplot variable.
                    # But if we try to plot these variables we may get an error.
                    continue

                # Store the data, and merge variables if that was requested.
                var_name = prefix + var + suffix
                to_merge = False
                # Merge if the variable has been saved already in the current group of files.
                # Also merge when the variable is already in tplot and merge is True.
                if (var_name in stored_variables) or (
                    var_name in pyspedas.tplot_tools.data_quants.keys() and (merge == True)
                ):
                    prev_data_quant = pyspedas.tplot_tools.data_quants[var_name]
                    to_merge = True

                tplot_data = {"x": unix_times, "y": this_masked_var}
                store_data(var_name, tplot_data)
                if var_name not in stored_variables:
                    stored_variables.append(var_name)

                if to_merge == True:
                    cur_data_quant = pyspedas.tplot_tools.data_quants[var_name]
                    plot_options = copy.deepcopy(pyspedas.tplot_tools.data_quants[var_name].attrs)
                    merged_data = [prev_data_quant, cur_data_quant]
                    pyspedas.tplot_tools.data_quants[var_name] = xr.concat(
                        merged_data, dim="time"
                    ).sortby("time")
                    pyspedas.tplot_tools.data_quants[var_name].attrs = plot_options

    # If we are interested in seeing a quick plot of the variables, do it
    if plot:
        tplot(stored_variables)

    return stored_variables



import os

CONFIG = {
    "local_data_dir": "geom_indices/",
    "gfz_remote_data_dir": "https://datapub.gfz-potsdam.de/download/10.5880.Kp.0001/Kp_definitive/",
    "noaa_remote_data_dir": "https://www.ngdc.noaa.gov/stp/space-weather/geomagnetic-data/INDICES/KP_AP/"
}

# override local data directory with environment variables
if os.environ.get("SPEDAS_DATA_DIR"):
    CONFIG["local_data_dir"] = os.sep.join(
        [os.environ["SPEDAS_DATA_DIR"], "geom_indices"]
    )

if os.environ.get("KP_DATA_DIR"):
    CONFIG["local_data_dir"] = os.environ["KP_DATA_DIR"]



import os
import logging
import warnings
from shutil import copyfileobj, copy
from tempfile import NamedTemporaryFile
from pyspedas.tplot_tools import time_double, time_string
from pyspedas.projects.mms.mms_login_lasp import mms_login_lasp
from pyspedas.projects.mms.mms_config import CONFIG
from pyspedas.projects.mms.mec_ascii.mms_get_local_ancillary_files import mms_get_local_ancillary_files
from pyspedas.projects.mms.mec_ascii.mms_load_qf_tplot import mms_load_qf_tplot

from pyspedas.utilities.download import is_fsspec_uri
import fsspec

def mms_get_tetrahedron_qf(trange=['2015-10-16', '2015-10-17'], no_download=False,
    suffix='', always_prompt=False):
    """
    Helper routine for loading tetrahedron QF data (ASCII files from the SDC); not meant to be called directly
    """

    local_data_dir = CONFIG['local_data_dir']
    download_only = CONFIG['download_only']

    start_time = time_double(trange[0])-60*60*24.
    end_time = time_double(trange[1])

    # check if end date is anything other than 00:00:00, if so
    # add a day to the end time to ensure that all data is downloaded
    if type(trange[1]) == str:
        endtime_day = time_double(time_string(time_double(trange[1]), fmt='%Y-%m-%d'))
    else:
        endtime_day = time_double(time_string(trange[1], fmt='%Y-%m-%d'))

    if end_time > endtime_day:
        add_day = 60*60*24.
    else:
        add_day = 0

    start_time_str = time_string(start_time, fmt='%Y-%m-%d')
    end_time_str = time_string(end_time+add_day, fmt='%Y-%m-%d')


    user = None
    if not no_download:
        sdc_session, user = mms_login_lasp(always_prompt=always_prompt)

    out_files = []

    sep = "/" if is_fsspec_uri(local_data_dir) else os.path.sep
    out_dir = sep.join([local_data_dir, 'ancillary', 'mms', 'tetrahedron_qf'])

    if CONFIG['no_download'] != True and no_download != True:
        dates_for_query = '&start_date='+start_time_str+'&end_date='+end_time_str

        if user is None:
            url = 'https://lasp.colorado.edu/mms/sdc/public/files/api/v1/file_info/ancillary?sc_id=mms&product=defq'+dates_for_query
        else:
            url = 'https://lasp.colorado.edu/mms/sdc/sitl/files/api/v1/file_info/ancillary?sc_id=mms&product=defq'+dates_for_query

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=ResourceWarning)
            http_request = sdc_session.get(url, verify=True)
            if http_request.status_code != 200:
                logging.warning("Request to MMS SDC returned HTTP status code %d",http_request.status_code)
                logging.warning("Text: %s", http_request.text)
                logging.warning("URL: %s", url)
                return
            else:
                http_json = http_request.json()

        files_in_interval = http_json['files']

        for file in files_in_interval:
            out_file = sep.join([out_dir, file['file_name']])

            if is_fsspec_uri(out_file):
                protocol, path = out_file.split("://")
                fs = fsspec.filesystem(protocol)

                exists = fs.exists(out_file)
            else:
                exists = os.path.exists(out_file)

            if exists:
                if is_fsspec_uri(out_file):
                    f_size = fs.size(out_file)
                else:
                    f_size = os.stat(out_file).st_size

                if str(f_size) == str(file['file_size']):
                    out_files.append(out_file)
                    http_request.close()
                    continue

            if user is None:
                download_url = 'https://lasp.colorado.edu/mms/sdc/public/files/api/v1/download/ancillary?file=' + file['file_name']
            else:
                download_url = 'https://lasp.colorado.edu/mms/sdc/sitl/files/api/v1/download/ancillary?file=' + file['file_name']

            logging.info('Downloading ' + file['file_name'] + ' to ' + out_dir)

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=ResourceWarning)
                fsrc = sdc_session.get(download_url, stream=True, verify=True)

            ftmp = NamedTemporaryFile(delete=False)

            with open(ftmp.name, 'wb') as f:
                copyfileobj(fsrc.raw, f)

            if is_fsspec_uri(out_dir):
                protocol, _ = out_dir.split("://")
                fs = fsspec.filesystem(protocol)

                fs.makedirs(out_dir, exist_ok=True)

                # if the download was successful, put at URI specified
                fs.put(ftmp.name, out_file)
            else:
                if not os.path.exists(out_dir):
                    os.makedirs(out_dir)

                # if the download was successful, copy to data directory
                copy(ftmp.name, out_file)

            out_files.append(out_file)
            fsrc.close()
            ftmp.close()

    if not no_download:
        sdc_session.close()

    if download_only:
        return

    # if no files are found remotely, try locally
    if len(out_files) == 0:
        out_files = mms_get_local_ancillary_files(trange=[start_time_str, end_time_str])

    return mms_load_qf_tplot(sorted(out_files), suffix=suffix, trange=trange)




import unittest
import numpy as np
from pyspedas.projects.mms.particles.mms_part_getspec import mms_part_getspec
from pyspedas.tplot_tools import data_exists, get_data, tplot_names, tplot, get_coords, get_units

global_display=True

class PGSTests(unittest.TestCase):
    def test_pgs_errors(self):
        # no trange specified
        mms_part_getspec()
        # unsupported instrument
        mms_part_getspec(trange=['2015-10-16/13:06:00', '2015-10-16/13:06:10'], instrument='feeps')

    def test_fpi_brst_fac_type(self):
        mms_part_getspec(trange=['2015-10-16/13:06:00', '2015-10-16/13:06:10'],
                         data_rate='brst',
                         species='i',
                         fac_type='phigeo',
                         output='pa gyro moments')
        self.assertTrue(data_exists('mms1_dis_dist_brst_density'))
        mms_part_getspec(trange=['2015-10-16/13:06:00', '2015-10-16/13:06:10'],
                         data_rate='brst',
                         species='i',
                         fac_type='xgse',
                         output='pa gyro moments')
        self.assertTrue(data_exists('mms1_dis_dist_brst_density'))

    def test_fpi_brst_i(self):
        mms_part_getspec(trange=['2015-10-16/13:06:00', '2015-10-16/13:06:10'],
                         data_rate='brst',
                         species='i',
                         output='energy theta phi pa gyro moments',
                         prefix='pre_')
        self.assertTrue(data_exists('pre_mms1_dis_dist_brst_density'))
        self.assertTrue(data_exists('pre_mms1_dis_dist_brst_velocity'))
        self.assertTrue(data_exists('pre_mms1_dis_dist_brst_avgtemp'))
        self.assertTrue(data_exists('pre_mms1_dis_dist_brst_energy'))
        self.assertTrue(data_exists('pre_mms1_dis_dist_brst_theta'))
        self.assertTrue(data_exists('pre_mms1_dis_dist_brst_phi'))
        self.assertTrue(data_exists('pre_mms1_dis_dist_brst_pa'))
        self.assertTrue(data_exists('pre_mms1_dis_dist_brst_gyro'))

    def test_fpi_mom_metadata(self):
        # Check coordinate systems and units for moments data, with and without sdc_units flag
        mms_part_getspec(trange=['2015-10-16/13:06:00', '2015-10-16/13:06:10'],
                         data_rate='brst',
                         species='i',
                         output='moments',
                         prefix='pre_',
                         )
        mms_part_getspec(trange=['2015-10-16/13:06:00', '2015-10-16/13:06:10'],
                         data_rate='brst',
                         species='i',
                         output='moments',
                         prefix='pre_',
                         suffix='_sdc_units',
                         sdc_units=True)

        self.assertTrue(data_exists('pre_mms1_dis_dist_brst_qflux'))
        self.assertTrue(data_exists('pre_mms1_dis_dist_brst_ptens'))
        self.assertTrue(data_exists('pre_mms1_dis_dist_brst_qflux_sdc_units'))
        self.assertTrue(data_exists('pre_mms1_dis_dist_brst_ptens_sdc_units'))

        tplot(['pre_mms1_dis_dist_brst_qflux','pre_mms1_dis_dist_brst_ptens',
               'pre_mms1_dis_dist_brst_qflux_sdc_units','pre_mms1_dis_dist_brst_ptens_sdc_units'])

        self.assertEqual(get_coords('pre_mms1_dis_dist_brst_qflux').lower(), 'dbcs')
        self.assertEqual(get_units('pre_mms1_dis_dist_brst_qflux'), 'eV/(cm^2-s)')
        self.assertEqual(get_coords('pre_mms1_dis_dist_brst_ptens').lower(), 'dbcs')
        self.assertEqual(get_units('pre_mms1_dis_dist_brst_ptens'), 'eV/cm^3')

        self.assertEqual(get_units('pre_mms1_dis_dist_brst_qflux_sdc_units'), 'mW/m^2')
        self.assertEqual(get_units('pre_mms1_dis_dist_brst_ptens_sdc_units'), 'nPa')

        # Check order of magnitude of qflux and ptens outputs, with/without the sdc_units flag,
        # to make sure the values were actually converted and not just the metadata updated.

        qfdat_std = get_data('pre_mms1_dis_dist_brst_qflux')
        qfdat_sdc = get_data('pre_mms1_dis_dist_brst_qflux_sdc_units')
        ptdat_std = get_data('pre_mms1_dis_dist_brst_ptens')
        ptdat_sdc = get_data('pre_mms1_dis_dist_brst_ptens_sdc_units')

        qfdat_std_max = np.abs(np.max(qfdat_std.y))
        qfdat_sdc_max = np.abs(np.max(qfdat_sdc.y))
        ptdat_std_max = np.abs(np.max(ptdat_std.y))
        ptdat_sdc_max = np.abs(np.max(ptdat_sdc.y))

        self.assertTrue(qfdat_std_max > 5.0e10)
        self.assertTrue(ptdat_std_max > 4000)
        self.assertTrue(qfdat_sdc_max < 1.0)
        self.assertTrue(ptdat_sdc_max < 1.0)

    def test_fpi_disable_pe_corr(self):
        mms_part_getspec(trange=['2015-10-16/13:06:07', '2015-10-16/13:06:08'],
                         data_rate='brst',
                         species='e',
                         output='energy',
                         disable_photoelectron_corrections=True,
                         prefix=None,
                         suffix=None)
        self.assertTrue(data_exists('mms1_des_dist_brst_energy'))

    def test_fpi_brst_e(self):
        mms_part_getspec(trange=['2015-10-16/13:06:07', '2015-10-16/13:06:08'],
                         data_rate='brst',
                         species='e',
                         output='energy theta phi pa gyro moments')
        self.assertTrue(data_exists('mms1_des_dist_brst_density'))
        self.assertTrue(data_exists('mms1_des_dist_brst_velocity'))
        self.assertTrue(data_exists('mms1_des_dist_brst_avgtemp'))
        self.assertTrue(data_exists('mms1_des_dist_brst_energy'))
        self.assertTrue(data_exists('mms1_des_dist_brst_theta'))
        self.assertTrue(data_exists('mms1_des_dist_brst_phi'))
        self.assertTrue(data_exists('mms1_des_dist_brst_pa'))
        self.assertTrue(data_exists('mms1_des_dist_brst_gyro'))

    def test_fpi_brst_limits(self):
        mms_part_getspec(trange=['2015-10-16/13:06:07', '2015-10-16/13:06:08'],
                         data_rate='brst',
                         species='e',
                         theta=[0, 90],
                         phi=[0, 100],
                         gyro=[0, 180],
                         pitch=[45, 75],
                         energy=[1000, 20000],
                         output='energy theta phi pa gyro moments')
        self.assertTrue(data_exists('mms1_des_dist_brst_fac_density'))
        self.assertTrue(data_exists('mms1_des_dist_brst_fac_velocity'))
        self.assertTrue(data_exists('mms1_des_dist_brst_fac_avgtemp'))
        self.assertTrue(data_exists('mms1_des_dist_brst_fac_energy'))
        self.assertTrue(data_exists('mms1_des_dist_brst_theta'))
        self.assertTrue(data_exists('mms1_des_dist_brst_phi'))
        self.assertTrue(data_exists('mms1_des_dist_brst_pa'))
        self.assertTrue(data_exists('mms1_des_dist_brst_gyro'))

    def test_hpca_srvy_hplus(self):
        mms_part_getspec(trange=['2015-10-16/13:05', '2015-10-16/13:10'],
                         #data_rate='brst',
                         species='hplus',
                         instrument='hpca',
                         output='energy theta phi pa gyro moments')
        self.assertTrue(data_exists('mms1_hpca_hplus_phase_space_density_density'))
        self.assertTrue(data_exists('mms1_hpca_hplus_phase_space_density_velocity'))
        self.assertTrue(data_exists('mms1_hpca_hplus_phase_space_density_avgtemp'))
        self.assertTrue(data_exists('mms1_hpca_hplus_phase_space_density_energy'))
        self.assertTrue(data_exists('mms1_hpca_hplus_phase_space_density_theta'))
        self.assertTrue(data_exists('mms1_hpca_hplus_phase_space_density_phi'))
        self.assertTrue(data_exists('mms1_hpca_hplus_phase_space_density_pa'))
        self.assertTrue(data_exists('mms1_hpca_hplus_phase_space_density_gyro'))

    def test_hpca_srvy_oplus(self):
        mms_part_getspec(trange=['2015-10-16/13:05', '2015-10-16/13:10'],
                         species='oplus',
                         instrument='hpca',
                         output='energy')
        self.assertTrue(data_exists('mms1_hpca_oplus_phase_space_density_energy'))

    def test_hpca_srvy_heplus(self):
        mms_part_getspec(trange=['2015-10-16/13:05', '2015-10-16/13:10'],
                         species='heplus',
                         instrument='hpca',
                         output='energy')
        self.assertTrue(data_exists('mms1_hpca_heplus_phase_space_density_energy'))

    def test_hpca_srvy_heplusplus(self):
        mms_part_getspec(trange=['2015-10-16/13:05', '2015-10-16/13:10'],
                         species='heplusplus',
                         instrument='hpca',
                         output='energy')
        self.assertTrue(data_exists('mms1_hpca_heplusplus_phase_space_density_energy'))

    def test_pitch_angle_limits(self):
        # Regression test for application of pitch angle limits to energy spectra
        # Previously, energy spectra were always computed in standard coordinates,
        # so any pitch angle or gyro limits were ignored.  Now, if 'energy' or 'moments'
        # are requested, and non-default pitch or gyro limits are specified, the outputs are
        # changed to 'fac_energy' or 'fac_moments', and pitch and gyro limits are applied.
        # Note that this changes the energy and moment output variable names!  (They will
        # now have a '_fac_' infix.  JWL 2024-08-15

        trange = ['2023-10-20/08:15:00', '2023-10-20/08:15:15']
        pitch_ranges = [[0, 30], [60, 120], [150, 180]]

        pitch_range = pitch_ranges[0]
        mms_part_getspec(trange=trange, instrument='fpi', data_rate='fast', probe=1, species='e', pitch=pitch_range,
                         output='energy', suffix='_pa_0_30')
        self.assertTrue(data_exists('mms1_des_dist_fast_fac_energy_pa_0_30'))
        data0 = get_data('mms1_des_dist_fast_fac_energy_pa_0_30')

        pitch_range = pitch_ranges[1]
        mms_part_getspec(trange=trange, instrument='fpi', data_rate='fast', probe=1, species='e', pitch=pitch_range,
                         output='energy', suffix='_pa_60_120')
        self.assertTrue(data_exists('mms1_des_dist_fast_fac_energy_pa_60_120'))
        data1 = get_data('mms1_des_dist_fast_fac_energy_pa_60_120')
        tplot(['mms1_des_dist_fast_fac_energy_pa_0_30', 'mms1_des_dist_fast_fac_energy_pa_60_120'], display=global_display, save_png='energy_pitch_limits.png')

        # Check that the pitch angle limits were actually applied
        delta = np.max(np.abs(data1.y - data0.y))
        self.assertTrue(delta > 0.0)

if __name__ == '__main__':
    unittest.main()



from .load import load

# This routine was originally in lanl/__init__.py.
def mpa(
    trange=["2004-10-31", "2004-11-01"],
    probe="l1",
    datatype="k0",
    prefix="",
    suffix="",
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    time_clip=False,
    force_download=False,
):
    """
    Load data from the LANL Magnetospheric Plasma Analyzer (MPA)

    Parameters
    ----------
    trange : list of str, optional
        Time range of interest with the format ['YYYY-MM-DD','YYYY-MM-DD'] or
        to specify more or less than a day ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss'].
        Defaults to ['2004-10-31', '2004-11-01'].
    probe : str, optional
        The probe to load data for. Defaults to 'a1'.
        Valid probes (with gaps for some dates): 'l0', 'l1', 'l4', 'l7', 'l9', 'a1', 'a2'.
    datatype : str, optional
        The datatype. Defaults to 'k0'.
        Valid datatypes: 'k0', 'h0'.
        Data for 'h0' datatype is available only for the 'mpa' instrument and only for a few days in 1998.
    prefix : str, optional
        The tplot variable names will be given this prefix. By default, no prefix is added.
        Defaults to ''.
    suffix : str, optional
        The tplot variable names will be given this suffix. Defaults to ''.
    get_support_data : bool, optional
        Data with an attribute "VAR_TYPE" with a value of "support_data" will be loaded into tplot.
        By default, only loads in data with a "VAR_TYPE" attribute of "data". Defaults to False.
    varformat : str, optional
        The file variable formats to load into tplot. Wildcard character "*" is accepted.
        By default, all variables are loaded in. Defaults to None.
    varnames : list of str, optional
        List of variable names to load. If not specified, all data variables are loaded. Defaults to [].
    downloadonly : bool, optional
        Set this flag to download the CDF files, but not load them into tplot variables. Defaults to False.
    notplot : bool, optional
        Return the data in hash tables instead of creating tplot variables. Defaults to False.
    no_update : bool, optional
        If set, only load data from your local cache. Defaults to False.
    time_clip : bool, optional
        Time clip the variables to exactly the range specified in the trange keyword. Defaults to False.
    force_download : bool
        Download file even if local version is more recent than server version. Defaults to False.

    Returns
    -------
    list
        List of tplot variables created.
    """


    tvars = load(
        instrument="mpa",
        trange=trange,
        datatype=datatype,
        probe=probe,
        prefix=prefix,
        suffix=suffix,
        get_support_data=get_support_data,
        varformat=varformat,
        varnames=varnames,
        downloadonly=downloadonly,
        notplot=notplot,
        time_clip=time_clip,
        no_update=no_update,
        force_download=force_download
    )

    if tvars is None or notplot or downloadonly:
        return tvars

    return tvars



import numpy as np


def mpause_t96(pd, xgsm=None, ygsm=None, zgsm=None):
    """
    Calculate the location of the magnetopause using the Tsyganenko 1996 (T96) model.

    Parameters
    ----------
        pd (float):
            Solar wind ram pressure in nanopascals.
        xgsm, ygsm, zgsm  (array-like):
            Coordinates of points in Earth radii (Re) to check whether they are inside or outside the magnetopause.

    Returns
    -------
        xmgnp, ymgnp, zmgnp (array-like):
            Locations of the magnetopause boundary in Earth radii (Re).
        id (array-like):
            Array flagging whether the given points are inside (1), outside (-1), or not checked (np.nan).
        distan (array-like):
            Distance from the given points to the nearest point on the magnetopause boundary.

    Notes
    -----
    The pressure-dependent magnetopause that is used in the T96_01 model
    (TSYGANENKO, JGR, V.100, P.5599, 1995; ESA SP-389, P.181, O;T. 1996)

    Similar to mpause_t96.pro in IDL SPEDAS.
    """

    # Constants defined in the T96 model
    a0 = 70.0
    s00 = 1.08
    x00 = 5.48
    pd0 = 2.0  # Average solar wind ram pressure in nPa

    # Adjust magnetopause parameters according to the actual solar wind pressure
    rat = pd / pd0
    rat16 = rat**0.14
    a = a0 / rat16
    s0 = s00
    x0 = x00 / rat16
    xm = x0 - a

    # Calculate the magnetopause boundary
    n = np.arange(1, 46)
    tau = 1 - 10 ** (-5) * n**3
    xmgnp_half = x0 - a * (1 - s0 * tau)
    arg = (s0**2 - 1) * (1 - tau**2)

    # Ensure argument of sqrt is non-negative
    rhomgnp = a * np.sqrt(np.maximum(arg, 0))
    ymgnp_half = rhomgnp
    xmgnp = np.concatenate([xmgnp_half[::-1], xmgnp_half])
    ymgnp = np.concatenate([-ymgnp_half[::-1], ymgnp_half])
    zmgnp = np.concatenate([-ymgnp_half[::-1], ymgnp_half])

    # Initialize output variables
    id = np.full(len(xgsm), np.nan) if xgsm is not None else None
    distan = np.full(len(xgsm), np.nan) if xgsm is not None else None
    xmp0 = np.full(len(xgsm), np.nan) if xgsm is not None else None
    ymp0 = np.full(len(xgsm), np.nan) if xgsm is not None else None
    zmp0 = np.full(len(xgsm), np.nan) if xgsm is not None else None

    # Flag points inside or outside the magnetopause
    if xgsm is not None and ygsm is not None and zgsm is not None:
        if len(xgsm) != len(ygsm) or len(ygsm) != len(zgsm):
            raise ValueError("Input XGSM, YGSM, ZGSM arrays must have the same length")

        phi = np.arctan2(ygsm, zgsm)
        rho = np.sqrt(ygsm**2 + zgsm**2)

        # Inside magnetosphere
        inside_indices = xgsm < xm
        if inside_indices.any():
            rhomgnp_boundary = a * np.sqrt(s0**2 - 1)
            index1_inside = rho[inside_indices] <= rhomgnp_boundary
            id[inside_indices] = np.where(index1_inside, 1, -1)

            xmp0[inside_indices] = xgsm[inside_indices]
            ymp0[inside_indices] = rhomgnp_boundary * np.sin(phi[inside_indices])
            zmp0[inside_indices] = rhomgnp_boundary * np.cos(phi[inside_indices])
            distan[inside_indices] = np.sqrt(
                (xgsm[inside_indices] - xmp0[inside_indices]) ** 2
                + (ygsm[inside_indices] - ymp0[inside_indices]) ** 2
                + (zgsm[inside_indices] - zmp0[inside_indices]) ** 2
            )

        # Outside magnetosphere
        outside_indices = xgsm >= xm
        if outside_indices.any():
            xksi = (xgsm[outside_indices] - x0) / a + 1.0
            xdzt = rho[outside_indices] / a
            sq1 = np.sqrt((1.0 + xksi) ** 2 + xdzt**2)
            sq2 = np.sqrt((1.0 - xksi) ** 2 + xdzt**2)
            sigma = 0.5 * (sq1 + sq2)
            tau = 0.5 * (sq1 - sq2)
            arg = (s0**2 - 1.0) * (1.0 - tau**2)
            arg[arg < 0] = 0  # Clip negative values to zero
            rhomgnp = a * np.sqrt(arg)

            id[outside_indices] = np.where(sigma >= s0, -1, 1)

            xmp0[outside_indices] = x0 - a * (1.0 - s0 * tau)
            ymp0[outside_indices] = rhomgnp * np.sin(phi[outside_indices])
            zmp0[outside_indices] = rhomgnp * np.cos(phi[outside_indices])
            distan[outside_indices] = np.sqrt(
                (xgsm[outside_indices] - xmp0[outside_indices]) ** 2
                + (ygsm[outside_indices] - ymp0[outside_indices]) ** 2
                + (zgsm[outside_indices] - zmp0[outside_indices]) ** 2
            )

    return xmgnp, ymgnp, zmgnp, id, distan



import numpy as np

from pyspedas.tplot_tools import get_data, store_data, options, clip, ylim

from ...satellite.erg.load import load
from ...satellite.erg.get_gatt_ror import get_gatt_ror

from typing import List, Union, Optional, Dict, Any

def gmag_isee_fluxgate(
    trange: List[str] = ['2020-08-01', '2020-08-02'],
    suffix: str = '',
    site: Union[str, List[str]] = 'all',
    datatype: Union[str, List[str]] = 'all',
    get_support_data: bool = False,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    ror: bool = True,
    force_download=False,
) -> Union[Dict, None, List[Union[str, Any]]]:
    """
    Load data from ISEE Fluxgate Magnetometers

    Parameters
    ----------
    trange: list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2020-08-01', '2020-08-02']

    suffix: str
            The tplot variable names will be given this suffix.  Default: ''

    site: str or list of str
            The site or list of sites to load.
            Valid values: 'msr', 'rik', 'kag', 'ktb', 'lcl', 'mdm', 'tew', 'all'
            Default: ['all']

    datatype: str or list of str
            The data types to load. Valid values: '64hz', '1sec', '1min', '1h', 'all'
            Default: 'all'

    get_support_data: bool
            If true, data with an attribute "VAR_TYPE" with a value of "support_data"
            or 'data' will be loaded into tplot. Default: False

    varformat: str
            The CDF file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  Default: None (all variables will be loaded).

    varnames: list of str
            List of variable names to load. Default: [] (all variables will be loaded)

    downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables. Default: False

    notplot: bool
            Return the data in hash tables instead of creating tplot variables. Default: False

    no_update: bool
            If set, only load data from your local cache. Default: False

    uname: str
            User name.  Default: None

    passwd: str
            Password. Default: None

    time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword. Default: False

    ror: bool
            If set, print PI info and rules of the road. Default: True

    force_download: bool
        Download file even if local version is more recent than server version
        Default: False

    Returns
    -------

    Examples
    ________

    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> fluxgate_vars=pyspedas.projects.erg.gmag_isee_fluxgate(trange=['2020-08-01','2020-08-02'], site='all')
    >>> tplot('isee_fluxgate_mag_ktb_1min_hdz')

    """
    site_code_all = ['msr', 'rik', 'kag', 'ktb', 'lcl', 'mdm', 'tew']
    tres_all=['64hz', '1sec', '1min', '1h']
    if isinstance(datatype, str):
        datatype = datatype.lower()
        datatype = datatype.split(' ')
    elif isinstance(datatype, list):
        for i in range(len(datatype)):
            datatype[i] = datatype[i].lower()

    if 'all' in datatype:
        datatype=tres_all
    datatype = list(set(datatype).intersection(tres_all))
    if len(datatype) < 1:
        return

    if '64' in datatype:
        index = np.where(np.array(datatype) == '64')[0][0]
        datatype[index] = '64hz'
    elif  '1s' in datatype:
        index = np.where(np.array(datatype) == '1s')[0][0]
        datatype[index] = '1sec'
    elif  '1m' in datatype:
        index = np.where(np.array(datatype) == '1m')[0][0]
        datatype[index] = '1min'
    elif  '1hr' in datatype:
        index = np.where(np.array(datatype) == '1hr')[0][0]
        datatype[index] = '1h'

    
    if isinstance(site, str):
        site_code = site.lower()
        site_code = site_code.split(' ')
    elif isinstance(site, list):
        site_code = []
        for i in range(len(site)):
            site_code.append(site[i].lower())
    if 'all' in site_code:
        site_code = site_code_all
    
    site_code = list(set(site_code).intersection(site_code_all))

    prefix = 'isee_fluxgate_'
    if notplot:
        loaded_data = {}
    else:
        loaded_data = []
    for site_input in site_code:
        for data_type_in in datatype:
            fres = data_type_in
            if fres == '64hz':
                file_res = 3600.
                pathformat = 'ground/geomag/isee/fluxgate/'+fres+'/'+site_input\
                                +'/%Y/%m/isee_fluxgate_'+fres+'_'+site_input+'_%Y%m%d%H_v??.cdf'
            if fres == '1h':
                fres = '1min'
            if (fres == '1sec') or (fres == '1min'):
                file_res = 3600. * 24
                pathformat = 'ground/geomag/isee/fluxgate/'+fres+'/'+site_input\
                                +'/%Y/isee_fluxgate_'+fres+'_'+site_input+'_%Y%m%d_v??.cdf'
            
            loaded_data_temp = load(pathformat=pathformat, file_res=file_res, trange=trange, datatype=datatype, prefix=prefix, suffix='_'+site_input+suffix, get_support_data=get_support_data,
                            varformat=varformat, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd, force_download=force_download)
            
            if notplot:
                loaded_data.update(loaded_data_temp)
            else:
                loaded_data += loaded_data_temp
            if (len(loaded_data_temp) > 0) and ror:
                try:
                    gatt = get_gatt_ror(downloadonly, loaded_data)
                    print('**************************************************************************')
                    print(gatt["Logical_source_description"])
                    print('')
                    print(f'Information about {gatt["Station_code"]}')
                    print('PI and Host PI(s):')
                    print(gatt["PI_name"])
                    print('')
                    print('Affiliations: ')
                    print(gatt["PI_affiliation"])
                    print('')
                    print('Rules of the Road for ISEE Fluxgate Data Use:')
                    for gatt_text in gatt["TEXT"]:
                        print(gatt_text)
                    print(f'{gatt["LINK_TEXT"]} {gatt["HTTP_LINK"]}')
                    print('**************************************************************************')
                except:
                    print('printing PI info and rules of the road was failed')
                
            if (not downloadonly) and (not notplot):
                if fres == '1min':
                    fres_list = ['1min', '1h']
                else:
                    fres_list = [fres]
                for fres_in in fres_list:
                    current_tplot_name = prefix+'hdz_'+fres_in+'_' + site_input+suffix
                    if current_tplot_name in loaded_data:
                        get_data_vars = get_data(current_tplot_name)
                        if get_data_vars is None:
                            store_data(current_tplot_name, delete=True)
                        else:
                            new_tplot_name = prefix+'mag_'+site_input+'_'+fres_in+'_hdz'+suffix
                            store_data(current_tplot_name, newname=new_tplot_name)
                            loaded_data.remove(current_tplot_name)
                            loaded_data.append(new_tplot_name)
                            clip(new_tplot_name, -1e+4, 1e+4)
                            get_data_vars = get_data(new_tplot_name)
                            ylim(new_tplot_name, np.nanmin(get_data_vars[1]), np.nanmax(get_data_vars[1]))
                            options(new_tplot_name, 'legend_names', ['H','D','Z'])
                            options(new_tplot_name, 'Color', ['b', 'g', 'r'])
                            options(new_tplot_name, 'ytitle', '\n'.join(new_tplot_name.split('_')))


    return loaded_data



"""Test gmag and themis load functions."""
import os
import unittest
import logging
import pyspedas
import pyspedas
from pyspedas.tplot_tools import data_exists, get_coords



class GmagTestCases(unittest.TestCase):
    """Test GMAG functions."""

    def test_get_group(self):
        """Get gmag stations of a group."""
        from pyspedas.projects.themis.ground.gmag import get_group
        self.assertTrue(get_group('ccnv') == ['epo'])

    def test_gmag_list(self):
        """Get gmag list of stations."""
        from pyspedas.projects.themis.ground.gmag import gmag_list
        l = gmag_list()
        for site in ['abk', 'akul', 'amd', 'amer', 'amk']:
            self.assertTrue(site in l)

    def test_gmag_groups(self):
        """Get gmag groups."""
        from pyspedas.projects.themis.ground.gmag import gmag_groups
        gmag_table = gmag_groups()
        keylist = list(gmag_table.keys())
        for group in ['kyoto', 'sgu', 'autx', 'ae', 'aari']:
            self.assertTrue(group in keylist)

    def test_check_gmag(self):
        """Check a gmag station."""
        from pyspedas.projects.themis.ground.gmag import check_gmag
        self.assertTrue(check_gmag('ccnv') == 1)
        self.assertTrue(check_gmag('abcd') == 0)

    def test_load_gmag_data(self):
        """Load gmag."""
        pyspedas.projects.themis.gmag(varnames=['thg_mag_amer'], sites='amer')
        self.assertTrue(data_exists('thg_mag_amer'))


class LoadTestCases(unittest.TestCase):
    """Test themis load functions."""

    def test_load_state_data(self):
        """Load state."""
        pyspedas.projects.themis.state(varnames=['thc_pos'])
        self.assertTrue(data_exists('thc_pos'))

    def test_load_state_support_data(self):
        """Load state."""
        state_vars = pyspedas.projects.themis.state(probe='a',trange=['2023-03-23','2023-03-24'], get_support_data=True)
        self.assertTrue('tha_spinras' in state_vars)
        self.assertTrue(data_exists('tha_spinras'))

    def test_load_state_support_data_multiprobe(self):
        """Load state."""
        state_vars = pyspedas.projects.themis.state(probe=['a','b'],trange=['2023-03-23','2023-03-24'], get_support_data=True)
        self.assertTrue('tha_spinras' in state_vars)
        self.assertTrue(data_exists('tha_spinras'))
        self.assertTrue('thb_spinras' in state_vars)
        self.assertTrue(data_exists('thb_spinras'))

    def test_load_sst_data(self):
        """Load SST."""
        pyspedas.projects.themis.sst(varnames=['thc_psif_en_eflux'])
        self.assertTrue(data_exists('thc_psif_en_eflux'))

    def test_load_fgm_data(self):
        """Load FGM."""
        pyspedas.projects.themis.fgm(varnames=['thc_fgs_btotal'])
        self.assertTrue(data_exists('thc_fgs_btotal'))

    def test_load_fit_data(self):
        """Load FIT."""
        pyspedas.projects.themis.fit(varnames=['thc_fgs_gse'])
        self.assertTrue(data_exists('thc_fgs_gse'))

    def test_load_esa_data(self):
        """Load ESA."""
        pyspedas.projects.themis.esa(varnames=['thc_peif_density'])
        self.assertTrue(data_exists('thc_peif_density'))

    def test_load_fft_data(self):
        """Load FFT."""
        pyspedas.projects.themis.fft(varnames=['thc_ffp_16_edc34'])
        self.assertTrue(data_exists('thc_ffp_16_edc34'))

    def test_load_fft_l1_data(self):
        """Load L1 FFT."""
        pyspedas.projects.themis.fft(level='l1', varnames=['thc_ffp_16'])
        self.assertTrue(data_exists('thc_ffp_16'))

    def test_load_fbk_data(self):
        """Load FBK."""
        pyspedas.projects.themis.fbk(varnames=['thc_fb_hff'])
        self.assertTrue(data_exists('thc_fb_hff'))

    def test_load_mom_data(self):
        """Load MOM."""
        pyspedas.projects.themis.mom(varnames=['thc_peim_density'])
        self.assertTrue(data_exists('thc_peim_density'))

    def test_load_gmom_data(self):
        """Load GMOM."""
        pyspedas.projects.themis.gmom(trange=['2020-01-01', '2020-01-01'],
                             varnames=['thc_ptiff_density'])
        self.assertTrue(data_exists('thc_ptiff_density'))

    def test_load_scm_data(self):
        """Load SCM."""
        pyspedas.projects.themis.scm(varnames=['thc_scf_btotal'])
        self.assertTrue(data_exists('thc_scf_btotal'))

    def test_load_scm_l1_data(self):
        """Load L1 SCM."""
        pyspedas.projects.themis.scm(level='l1', varnames=['thc_scf'])
        self.assertTrue(data_exists('thc_scf'))

    def test_load_efi_l1_data(self):
        """Load L1 EFI."""
        vars = pyspedas.projects.themis.efi(level='l1', varnames=['thc_eff','thc_efp'])
        self.assertTrue(data_exists('thc_eff'))
        self.assertTrue(data_exists('thc_efp'))
        self.assertFalse('thc_efw' in vars)

    def test_load_efi_l1_datatype(self):
        """Load L1 EFI."""
        vars = pyspedas.projects.themis.efi(level='l1', datatype=['eff', 'efp'])
        self.assertTrue(data_exists('thc_eff'))
        self.assertTrue(data_exists('thc_efp'))
        self.assertFalse('thc_efw' in vars)

    def test_load_efi_data(self):
        """Load EFI."""
        vars = pyspedas.projects.themis.efi(time_clip=True, varnames=['thc_eff_e12_efs'])
        self.assertTrue(data_exists('thc_eff_e12_efs'))
        self.assertFalse('thc_efw_gse' in vars)

    def test_load_l2_efw_data(self):
        """Load EFI L2 wave burst data"""
        vars = pyspedas.projects.themis.efi(trange=['2017-01-01','2017-01-02'], probe='a',time_clip=True, datatype='efw')
        self.assertTrue(data_exists('tha_efw_gse'))

    def test_load_l2_efp_data(self):
        """Load EFI L2 particle burst data"""
        vars = pyspedas.projects.themis.efi(trange=['2017-01-01','2017-01-02'], probe='a',time_clip=True, datatype='efp')
        self.assertTrue(data_exists('tha_efp_gse'))

    def test_load_slp_data(self):
        pyspedas.projects.themis.slp()
        # Check that all data is loaded
        self.assertTrue(data_exists('slp_sun_ltime'))
        self.assertTrue(data_exists('slp_lun_ltime'))
        self.assertTrue(data_exists('slp_sun_pos'))
        self.assertTrue(data_exists('slp_lun_vel'))
        self.assertTrue(data_exists('slp_sun_ltime'))
        self.assertTrue(data_exists('slp_lun_ltime'))
        self.assertTrue(data_exists('slp_lun_vel'))
        self.assertTrue(data_exists('slp_sun_att_x'))
        self.assertTrue(data_exists('slp_sun_att_z'))
        self.assertTrue(data_exists('slp_lun_att_x'))
        self.assertTrue(data_exists('slp_lun_att_z'))
        # Check that coordinate systems are set properly
        self.assertEqual(get_coords('slp_sun_pos').lower(),'gei')
        self.assertEqual(get_coords('slp_sun_vel').lower(),'gei')
        self.assertEqual(get_coords('slp_sun_att_x').lower(),'gei')
        self.assertEqual(get_coords('slp_sun_att_z').lower(),'gei')
        self.assertEqual(get_coords('slp_lun_pos').lower(),'gei')
        self.assertEqual(get_coords('slp_lun_vel').lower(),'gei')
        self.assertEqual(get_coords('slp_lun_att_x').lower(),'gei')
        self.assertEqual(get_coords('slp_lun_att_z').lower(),'gei')

    def test_downloadonly(self):
        """Downloadonly keyword."""
        files = pyspedas.projects.themis.efi(downloadonly=True,
                                    trange=['2014-2-15', '2014-2-16'],
                                    varnames=['thc_eff_e12_efs'])
        self.assertTrue(os.path.exists(files[0]))


if __name__ == '__main__':
    unittest.main()



"""
Python implementation of IDL SPEDAS wav_data.pro.

Uses wavelet2.py which is a wrapper for wavelet98.py.
"""

import numpy as np
import logging
from pyspedas.analysis.wavelet2 import wavelet2
from pyspedas.analysis.reduce_tres import reduce_tres
from pyspedas.analysis.roundsig import roundsig
from pyspedas.analysis.interp_gap import interp_gap
from scipy.ndimage import uniform_filter1d
from pyspedas import get_data, store_data, options, tnames


def smooth_wavelet(wv, wid, gaussian=False):
    """
    Smooth a 3D wavelet array along the first axis using a Gaussian or boxcar kernel.

    Implementation of IDL smooth_wavelet procedure to match the exact behavior.

    Parameters
    ----------
    wv : ndarray
        Input 3D array of shape (nt, nj, nk), to be smoothed along axis 0.
        Modified in-place to match IDL behavior.
    wid : array_like
        1D array specifying smoothing width for each scale (length nj).
    gaussian : bool, optional
        If True, use a Gaussian kernel. Otherwise, use IDL's smooth function equivalent.

    Notes
    -----
    This function modifies wv in-place to match IDL behavior.
    It does not return a new array.
    """
    dim = list(wv.shape) + [1, 1]
    nk = dim[2] if len(dim) > 2 else 1
    nj = dim[1] if len(dim) > 1 else 1
    nt = dim[0]

    wid = np.asarray(wid)

    if gaussian:
        # Calculate widi array for Gaussian case
        widi = np.round(wid / 2).astype(int) * 2 + 1
        # IDL behavior: width <= 1 means no smoothing
        widi = np.clip(widi, 1, nt - 1)

        for k in range(nk):
            for j in range(nj):
                if widi[j] > 1:  # Only smooth if width > 1
                    # Create Gaussian kernel using dgen equivalent
                    kernel_size = widi[j]
                    vx = np.linspace(-2, 2, kernel_size)
                    kernel = np.exp(-(vx**2))
                    kernel = kernel / np.sum(kernel)

                    # Apply convolution with edge truncation
                    signal = (
                        wv[:, j, k]
                        if wv.ndim >= 3
                        else wv[:, j] if wv.ndim >= 2 else wv[:]
                    )
                    # Use numpy.convolve with 'same' mode and handle edges
                    convolved = np.convolve(signal, kernel, mode="same")

                    if wv.ndim >= 3:
                        wv[:, j, k] = convolved
                    elif wv.ndim >= 2:
                        wv[:, j] = convolved
                    else:
                        wv[:] = convolved
                # If widi[j] <= 1, leave the data unchanged (IDL behavior)
        return

    # Non-Gaussian case: recalculate widi to match IDL behavior
    widi = np.round(wid).astype(int)
    # IDL smooth() with width <= 1 returns original data unchanged
    widi = np.clip(widi, 1, nt - 1)

    for k in range(nk):
        for j in range(nj):
            if widi[j] > 1:  # Only smooth if width > 1
                # Use scipy's uniform_filter1d to match IDL's smooth function

                signal = (
                    wv[:, j, k] if wv.ndim >= 3 else wv[:, j] if wv.ndim >= 2 else wv[:]
                )

                # Apply uniform filter with size widi[j], handling NaNs and edge truncation
                smoothed = uniform_filter1d(
                    signal.astype(float),
                    size=min(widi[j], nt - 1),
                    mode="nearest",  # Equivalent to IDL's /edge_truncate
                )
                # smoothed = smooth(signal, width=widi[j], preserve_nans=True)

                if wv.ndim >= 3:
                    wv[:, j, k] = smoothed
                elif wv.ndim >= 2:
                    wv[:, j] = smoothed
                else:
                    wv[:] = smoothed
            # If widi[j] <= 1, leave the data unchanged (IDL behavior)


def cross_corr_wavelet(wa, wb, wid, gaussian=False):
    """
    Cross-correlation wavelet analysis.

    Implementation of IDL cross_corr_wavelet procedure.

    Parameters
    ----------
    wa : ndarray
        First wavelet array (complex)
    wb : ndarray
        Second wavelet array (complex)
    wid : array_like
        Smoothing width array
    gaussian : bool, optional
        Use Gaussian smoothing (default=False)

    Returns
    -------
    gam, coinc, quad, crat : tuple

        gam : ndarray
            Coherence array
        coinc : ndarray
            Coincidence (real part) array
        quad : ndarray
            Quadrature (imaginary part) array
        crat : ndarray
            Coincidence ratio array
    """
    # Cross-correlation
    wa = np.asarray(wa, dtype=complex)
    wb = np.asarray(wb, dtype=complex)
    p = wa * np.conj(wb)
    coinc = np.real(p)
    quad = np.imag(p)

    # Smooth the cross-correlation components
    smooth_wavelet(coinc, wid, gaussian=gaussian)
    smooth_wavelet(quad, wid, gaussian=gaussian)

    # Power calculations
    p = np.abs(wa) ** 2
    smooth_wavelet(p, wid, gaussian=gaussian)

    gam = np.abs(wb) ** 2
    smooth_wavelet(gam, wid, gaussian=gaussian)

    # Calculate coincidence ratio
    crat = coinc / gam

    # Calculate coherence
    p_gam = p * gam
    gam = (coinc**2 + quad**2) / p_gam

    # Normalize coincidence and quadrature
    p_sqrt = np.sqrt(p_gam)
    coinc = coinc / p_sqrt
    quad = quad / p_sqrt

    return gam, coinc, quad, crat


def rotate_wavelet2(wv, bfield, xdir=None, wid=None, rotmats=None, get_rotmats=False):
    """
    Rotate a 3D wavelet vector field into local coordinate frames defined by B and xdir.

    Implementation of IDL rotate_wavelet2 function.

    Parameters
    ----------
    wv : ndarray of shape (nt, jv+1, 3)
        Input wavelet transform data with 3 spatial components.

    bfield : ndarray of shape (nt, 3)
        Vector field that defines the "z" axis of the local coordinate system at each time.

    xdir : array-like of shape (3,), optional
        A reference direction (default: [0, 0, 1]) used to define the "y" axis with bfield.
        This should not be parallel to bfield.

    wid : array-like of shape (jv+1,), required if rotmats is not supplied
        Smoothing window width (used on field B) for each scale.

    rotmats : ndarray of shape (nt, jv+1, 3, 3), optional
        Precomputed rotation matrices.

    get_rotmats : bool, optional
        If True, return the computed rotation matrices (useful for reuse).

    Returns
    -------
    wvp, rotmats_out : tuple

        wvp : ndarray of shape (nt, jv+1, 3)
            Wavelet transform rotated into local coordinates.

        rotmats_out : ndarray
            If True, return the computed rotation matrices. Shape (nt, jv+1, 3, 3).
            If False, return None.
    """
    # Get dimensions like IDL
    dim = list(wv.shape)
    nt = dim[0]
    jv = dim[1] - 1  # IDL uses jv = dim[1] - 1

    if dim[2] != 3:
        raise ValueError("Must have 3 dimensions")

    # Set default xdir
    if xdir is None or len(xdir) != 3:
        xdir = np.array([0.0, 0.0, 1.0])
    else:
        xdir = np.asarray(xdir)

    # Initialize rotation matrix and output wavelet
    rot = np.zeros((nt, 3, 3), dtype=np.float32)
    wvp = np.zeros_like(wv, dtype=wv.dtype)

    # Check if using precomputed rotmats
    use_rotmat = (rotmats is not None) and (not get_rotmats)

    # Initialize rotmats for output if requested
    if get_rotmats:
        rotmats_out = np.zeros((nt, dim[1], 3, 3), dtype=np.float32)

    # Main loop over scales (j=0 to jv, inclusive)
    for j in range(jv + 1):
        if use_rotmat:
            # Use precomputed rotation matrices
            rot = rotmats[:, j, :, :]
        else:
            if wid is None:
                raise ValueError("wid must be provided if rotmats are not used.")

            # Smooth B field components
            for k in range(3):
                vwidth = min(int(wid[j]), nt - 1)
                if vwidth > 1:
                    # Use uniform_filter1d to match IDL's smooth with /edge_truncate
                    rot[:, k, 2] = uniform_filter1d(
                        bfield[:, k].astype(float), size=vwidth, mode="nearest"
                    )
                else:
                    rot[:, k, 2] = bfield[:, k]

            # Normalize z-axis (rot[*,*,2])
            z_norm = np.sqrt(np.sum(rot[:, :, 2] ** 2, axis=1))
            # Avoid division by zero
            z_norm = np.where(z_norm == 0, 1, z_norm)
            rot[:, :, 2] = rot[:, :, 2] / z_norm[:, np.newaxis]

            # Compute y-axis
            for ni in range(nt):
                rot[ni, :, 1] = np.cross(rot[ni, :, 2], xdir)

            # Normalize y-axis
            y_norm = np.sqrt(np.sum(rot[:, :, 1] ** 2, axis=1))
            y_norm = np.where(y_norm == 0, 1, y_norm)
            rot[:, :, 1] = rot[:, :, 1] / y_norm[:, np.newaxis]

            # Compute x-axis
            for ni in range(nt):
                rot[ni, :, 0] = np.cross(rot[ni, :, 1], rot[ni, :, 2])

            # Store rotation matrices if requested
            if get_rotmats:
                rotmats_out[:, j, :, :] = rot

        # Apply rotation to wavelet components
        wvp[:, j, 2] = (
            wv[:, j, 0] * rot[:, 0, 2]
            + wv[:, j, 1] * rot[:, 1, 2]
            + wv[:, j, 2] * rot[:, 2, 2]
        )
        wvp[:, j, 1] = (
            wv[:, j, 0] * rot[:, 0, 1]
            + wv[:, j, 1] * rot[:, 1, 1]
            + wv[:, j, 2] * rot[:, 2, 1]
        )
        wvp[:, j, 0] = (
            wv[:, j, 0] * rot[:, 0, 0]
            + wv[:, j, 1] * rot[:, 1, 0]
            + wv[:, j, 2] * rot[:, 2, 0]
        )

    if get_rotmats:
        return wvp, rotmats_out
    else:
        return wvp, None


def wav_data_set_options(varname, vdict):
    """Set pyspedas options for wavelet data visualization."""
    for key, value in vdict.items():
        options(varname, key, value)


def wav_data(
    varname,
    period=None,
    prange=None,
    frange=None,
    param=None,
    avg_period=6.0,
    nmorlet=None,
    tplot_prefix=None,
    magrat=False,
    per_axis=False,
    kolom=None,
    normval=None,
    normconst=1.0,
    get_components=False,
    tint_pow=None,
    fraction=False,
    rotmats=None,
    get_rotmats=False,
    wid=None,
    dimennum=None,
    rotate_pow=False,
    maxpoints=32768,  # 2^15 like IDL
    rbin=None,
    cross1=False,
    cross2=False,
    trange=None,
    resolution=0,  # Add resolution parameter like IDL
):
    """
    Python implementation of wav_data.pro. Computes wavelet transform of tplot variable,
    using the Torrence & Compo (1998) algorithm.

    Stores results in pytplot variables with normalization, masking, smoothing, and options.

    This function calls wavelet2.pro which sets various parameters for the wavelet transform in wavelet98.py.

    Parameters
    ----------
    varname : str
        Name of tplot variable to analyze
    period : array-like, optional
        Array of periods to use for wavelet analysis
    prange : array-like, optional
        Period range [min_period, max_period] in same units as time
    frange : array-like, optional
        Frequency range [min_freq, max_freq] - converted to prange
    param : float, optional
        Wavelet parameter (default=6 for Morlet wavelet)
    avg_period : float, optional
        Averaging period for smoothing (default=6.0)
    nmorlet : float, optional
        Number of Morlet wavelets, converted to param
    tplot_prefix : str, optional
        Prefix for output tplot variable names
    magrat : bool, optional
        Compute magnitude ratio analysis
    per_axis : bool, optional
        Use period (True) or frequency (False) for y-axis
    kolom : array-like, optional
        Kolmogorov spectrum for normalization
    normval : array-like or float, optional
        Normalization values
    normconst : float, optional
        Normalization constant (default=1.0)
    get_components : bool, optional
        Store individual component power spectra
    tint_pow : float, optional
        Power law for frequency normalization
    fraction : bool, optional
        Normalize by total power fraction
    rotmats : array-like, optional
        Precomputed rotation matrices for rotate_pow
    get_rotmats : bool, optional
        Return rotation matrices
    wid : array-like, optional
        Smoothing width array
    dimennum : int, optional
        Select specific component of input variable
    rotate_pow : bool, optional
        Perform field-aligned coordinate analysis
    maxpoints : int, optional
        Maximum number of time points (default=32768)
    rbin : int, optional
        Rebinning factor
    cross1 : bool, optional
        Cross-correlation analysis type 1
    cross2 : bool, optional
        Cross-correlation analysis type 2
    trange : array-like, optional
        Time range for analysis
    resolution : int, optional
        Time resolution reduction factor. If > 0, reduces time resolution
        of output data using reduce_tres function (default=0, no reduction)

    Returns
    -------
    dict
        Dictionary with the following keys:

            'wave': ndarray, wavelet transform data
            'power': ndarray, power spectral data
            'period': ndarray, period/frequency scale values
            'yax': ndarray, y-axis values (period or frequency)
            'time': ndarray, time values
            'returned_rotmats': ndarray or None, rotation matrices if get_rotmats=True and rotate_pow=True
    """
    # Get tplot variable name
    name = tplot_prefix if tplot_prefix else varname
    if tnames(name) is None:
        logging.info(f"Variable: {name} not found.")
        return

    # Initialize output
    returned_rotmats = None

    # Handle nmorlet parameter
    if nmorlet is not None:
        param = nmorlet * 2 * np.pi

    # Get data
    d = get_data(name)
    time = d[0]
    bfield = d[1]
    if time is None or len(time) == 0:
        logging.info(f"Variable: {name} has no data.")
        return

    # Select component if requested
    if dimennum is not None:
        bfield = bfield[:, dimennum]
        name = f"{name}({dimennum})"

    # Frequency/period range
    if frange is not None and len(frange) == 2:
        prange = [1.0 / frange[1], 1.0 / frange[0]]

    # Rebin if requested
    if rbin is not None:
        nn = len(time)
        nrbin = nn // rbin
        time = time[: nrbin * rbin].reshape(nrbin, rbin).mean(axis=1)
        bfield = bfield[: nrbin * rbin]
        if bfield.ndim == 2:
            bfield = bfield.reshape(nrbin, rbin, bfield.shape[1]).mean(axis=1)
        else:
            bfield = bfield.reshape(nrbin, rbin).mean(axis=1)

    # Limit maxpoints
    if len(time) > maxpoints and trange is None:
        msg = f"Too many time samples, (Pts:{len(time)},Limit:{maxpoints}). Please select a different time range"
        logging.info(msg)
        return

    # Apply trange
    if trange is not None:
        if len(trange) == 1:
            idxn = np.argmin(np.abs(time - trange[0]))
            rr = np.arange(
                max(0, idxn - maxpoints // 2), min(len(time), idxn + maxpoints // 2)
            )
            time = time[rr[0] : rr[-1] + 1]
            bfield = bfield[rr[0] : rr[-1] + 1]
        else:
            t0, t1 = min(trange), max(trange)
            w = np.where((time >= t0) & (time <= t1))[0]
            if len(w) == 0:
                logging.info("No data in that time range")
                return
            if len(w) > maxpoints:
                logging.info(
                    f"Too many time samples. (Pts:{len(w)},Limit:{maxpoints}). Please select a different time range"
                )
                return
            time = time[w]
            bfield = bfield[w]

    # dt and check sampling like IDL
    dtime = np.diff(time)
    dt = np.mean(dtime)

    # Check for uniform sampling and resample if needed
    tgap_threshold = 0.1
    # Fix resampling threshold to match IDL: total(abs(minmax(dtime)/dt-1)) gt tgap_threshold
    if (
        np.sum(np.abs(np.array([np.min(dtime), np.max(dtime)]) / dt - 1))
        > tgap_threshold
    ):
        resample = True
    else:
        resample = False

    if resample:
        # This part fixes any time axis non-uniformity by resampling onto a uniform grid
        logging.warning("Warning!!! Resampling data onto a uniform period")
        dsample = np.round(dtime / np.median(dtime)).astype(int)
        samples = np.concatenate(([0], np.cumsum(dsample, dtype=int)))
        grid_len = samples[-1] + 1

        re_time = np.full(grid_len, np.nan, dtype=float)
        re_time[samples] = time
        time = interp_gap(np.arange(grid_len, dtype=float), re_time)["y"]

        # Initialize re_B with the appropriate shape and data type
        if bfield.ndim == 1:
            re_b = np.full(grid_len, np.nan, dtype=bfield.dtype)
            re_b[samples] = bfield
        elif bfield.ndim == 2:
            re_b = np.full((grid_len, bfield.shape[1]), np.nan, dtype=bfield.dtype)
            re_b[samples, :] = bfield
        elif bfield.ndim == 3:
            re_b = np.full(
                (grid_len, bfield.shape[1], bfield.shape[2]), np.nan, dtype=bfield.dtype
            )
            re_b[samples, :, :] = bfield
        else:
            re_b = np.full((grid_len,) + bfield.shape[1:], np.nan, dtype=bfield.dtype)
            src = (slice(None),) * (bfield.ndim - 1)
            re_b[(samples,) + src] = bfield

        bfield = re_b

        dtime = np.diff(time)
        dt = np.mean(dtime)

    # Interpolate gaps in bfield to match IDL
    interp_out = interp_gap(time, bfield)
    nbad_count = interp_out["count"]
    if nbad_count > 0:
        logging.info(f"Warning!!! Interpolating {nbad_count} bad points in data")
        bfield = interp_out["y"]
        # bad_index = interp_out["index"] # not used

    # Pad for FFT
    pad = 2

    # Compute wavelet
    wave, period = wavelet2(
        bfield, dt, pad=pad, period=period, prange=prange, param=param
    )

    if isinstance(wave, int) and wave == -1:
        logging.info("Time interval too short, Returning")
        return

    nt, jv1 = wave.shape[:2]
    nk = wave.shape[2] if wave.ndim == 3 else 1

    # Magnitude wavelet
    wvmag = None
    if magrat:
        logging.info("Computing magnitude now")
        wvmag, _ = wavelet2(
            np.sqrt(np.sum(bfield**2, axis=-1)),
            dt,
            pad=pad,
            period=period,
            prange=prange,
            param=param,
        )

    # Axis labels
    yax = period if per_axis else 1.0 / period
    ysubtitle = "Seconds" if per_axis else "f (Hz)"
    mm = [np.min(yax), np.max(yax)]

    # Width used in smoothing
    if wid is None:
        wid = (period / 2 / dt * avg_period) * 2 + 1
    wid = np.maximum(wid, 3).astype(int)

    # Normalization
    normpow = normconst
    if normval is not None:
        normval_arr = np.asarray(normval, dtype=float)
        if normval_arr.size == nt:
            logging.info("Calculating Normalization")
            normpow = normval_arr[:, None] * np.full((1, jv1), normconst, dtype=float)
            smooth_wavelet(normpow, wid)
        elif normval_arr.size == 1:
            normpow = float(normval_arr.ravel()[0])

    tsfx = ""
    if fraction:
        # Fractional normalization
        logging.info("Calculating Fractional Power")
        if nk > 1:
            normpow = (np.sum(bfield**2, axis=-1))[:, None] * np.ones(jv1)
        else:
            normpow = (bfield**2)[:, None] * np.ones(jv1)
        smooth_wavelet(normpow, wid)
        normpow = 1.0 / normpow
        tsfx += "$<B^2>$"

    zrangetmp = None  # temp zrange
    if kolom is not None:
        # Kolmogorov normalization
        logging.info("Calculating Kolmogorov Normalization")
        if np.isscalar(normpow):
            normpow = normpow / (kolom * period ** (5.0 / 3.0))
        else:
            normpow = normpow / (np.ones(nt)[:, None] * (kolom * period ** (5.0 / 3.0)))
        tsfx += "$P_{K}$"
        zrangetmp = [0.1, 10.0]
    elif tint_pow is not None:
        # Power-law normalization
        logging.info("Calculating Power-Law Normalization")
        if np.isscalar(normpow):
            normpow = normpow / (period**tint_pow)
        else:
            normpow = normpow / (np.ones(nt)[:, None] * (period**tint_pow))
        tsfx += "*f"
        if tint_pow != 1:
            vsub = f"{tint_pow:4.2f}"
            tsfx += f"$_{{{vsub}}}$"
        zrangetmp = [0.0001, 1.0]

    # Apply normalization - handle scalar and array cases
    if np.isscalar(normpow):
        wave = wave * np.sqrt(normpow)
        if wvmag is not None:
            wvmag = wvmag * np.sqrt(normpow)
    else:
        for k in range(nk):
            wave[..., k] *= np.sqrt(normpow)
        if wvmag is not None:
            wvmag *= np.sqrt(normpow)

    # Power
    if nk == 1:
        wpow = np.abs(wave) ** 2
    else:
        wpow = np.sum(np.abs(wave) ** 2, axis=2)

    # zrange
    mask_final = 1.0
    zrange = [-1.0, 1.0]
    if zrangetmp is not None:
        zrange = zrangetmp
    else:
        zrange = roundsig(
            10 ** np.nanmean(np.log10(wpow * mask_final)), sigfig=0.2
        ) * np.array([0.1, 10.0])

    # Store main power with nmorlet suffix
    wvs = "_wv"
    if nmorlet is not None:
        wvs += str(int(nmorlet))

    # Use reduce_tres
    r = resolution if resolution else 0
    rdtime = reduce_tres(time, r)
    ztitle = ""  # ztitle changes for each tplot variable

    polopts = {
        "spec": 1,
        "yrange": mm,
        "ylog": 1,
        "ystyle": 1,
        "no_interp": 1,
        "zrange": [-1.0, 1.0],
        "zlog": 0,
        "zstyle": 1,
        "ztitle": ztitle,
        "ysubtitle": ysubtitle,
    }

    powopts = {
        "spec": 1,
        "yrange": mm,
        "ylog": 1,
        "ystyle": 1,
        "no_interp": 1,
        "zlog": 1,
        "zstyle": 1,
        "zrange": zrange,
        "ztitle": ztitle,
        "ysubtitle": ysubtitle,
    }

    # Apply mask
    powname = name + wvs + "_pow"
    reduced_pow = reduce_tres(wpow * mask_final, r)
    powopts["ztitle"] = "$P_{Tot}$" + tsfx
    store_data(
        powname, data={"x": rdtime, "y": reduced_pow, "v": yax}, attr_dict=powopts
    )
    wav_data_set_options(powname, powopts)

    if nk == 1:
        # Single component ends here
        return {"wave": wave, "power": wpow, "period": period, "yax": yax, "time": time}

    # Store components if requested
    if get_components and nk > 1:
        cstr = ["x", "y", "z", "4"]
        for ni in range(nk):
            powi = np.abs(wave[..., ni]) ** 2
            nistr = cstr[ni]
            varname = name + "_" + nistr + wvs + "_pow"
            powopts["ztitle"] = "$P_{" + nistr + "}$" + tsfx
            store_data(
                varname,
                data={"x": rdtime, "y": reduce_tres(powi * mask_final, r), "v": yax},
                attr_dict=powopts,
            )
            wav_data_set_options(varname, powopts)

    # Magnitude ratio
    if magrat and wvmag is not None:
        varname = name + wvs + "_rat_par"
        powb = np.abs(wvmag) ** 2
        pol = powb / wpow
        smooth_wavelet(pol, wid)
        ztitle = "$P_{mag}/P_{tot}$"
        ratopts = polopts.copy()
        ratopts["ztitle"] = ztitle
        ratopts["zrange"] = [0.0, 1.0]
        store_data(
            varname,
            data={"x": rdtime, "y": reduce_tres(pol * mask_final, r), "v": yax},
            attr_dict=ratopts,
        )
        wav_data_set_options(varname, ratopts)

    ratopts = polopts.copy()
    ratopts["ztitle"] = "$P_{\\parallel}/P_{tot}$"
    ratopts["zrange"] = [0, 1]

    if nk == 2:
        # Form right/left circular powers powr / powl
        powr = np.abs(wave[..., 0] + 1j * wave[..., 1]) ** 2
        powl = np.abs(wave[..., 0] - 1j * wave[..., 1]) ** 2
        pol = (powr - powl) / (powl + powr)
        smooth_wavelet(pol, wid)
        reduced_pol = reduce_tres(pol * mask_final, r)
        varname = name + wvs + "_pol_perp"
        polopts["ztitle"] = "$P_{\\perp}$"
        store_data(
            varname,
            data={"x": rdtime, "y": reduced_pol, "v": yax},
            attr_dict=polopts,
        )
        wav_data_set_options(varname, polopts)

    # Rotate wavelet
    if rotate_pow:
        logging.info("Computing power and polarization for TPLOT")
        wv_rot, returned_rotmats = rotate_wavelet2(
            wave, bfield, wid=wid, rotmats=rotmats, get_rotmats=get_rotmats
        )
        powr = np.abs(wv_rot[..., 0] + 1j * wv_rot[..., 1]) ** 2
        powl = np.abs(wv_rot[..., 0] - 1j * wv_rot[..., 1]) ** 2
        powb = np.abs(wv_rot[..., 2]) ** 2

        pol = powb / (powb + powl + powr)
        smooth_wavelet(pol, wid)
        varname = name + wvs + "_pol_par"
        store_data(
            varname,
            data={"x": rdtime, "y": reduce_tres(pol * mask_final, r), "v": yax},
            attr_dict=ratopts,
        )
        wav_data_set_options(varname, ratopts)

        pol = (powr - powl) / (powl + powr)
        smooth_wavelet(pol, wid)
        varname = name + wvs + "_pol_perp"
        polopts["ztitle"] = "$P_{\\perp}$"
        store_data(
            varname,
            data={"x": rdtime, "y": reduce_tres(pol * mask_final, r), "v": yax},
            attr_dict=polopts,
        )
        wav_data_set_options(varname, polopts)

    # Hermitian analysis - hermition_k
    # Removed because IDL code uses the function 'polyroots' which is not defined

    if cross1 and nk >= 2:
        logging.info("Computing cross-correlation cross1")
        # Linear polarization cross-correlation
        wa = wave[..., 0] + 1j * wave[..., 1]
        wb = wave[..., 0] - 1j * wave[..., 1]

        cpol, cpow, cpowb, _ = cross_corr_wavelet(wa, wb, wid, gaussian=True)

        vname = name + wvs + "_gam_lin"
        ratopts["ztitle"] = "$g_l$"
        store_data(
            vname,
            data={"x": rdtime, "y": reduce_tres(cpol * mask_final, r), "v": yax},
            attr_dict=ratopts,
        )
        wav_data_set_options(vname, ratopts)

        vname = name + wvs + "_coin_lin"
        polopts["ztitle"] = "$C_l$"
        store_data(
            vname,
            data={"x": rdtime, "y": reduce_tres(cpow * mask_final, r), "v": yax},
            attr_dict=polopts,
        )
        wav_data_set_options(vname, polopts)

        vname = name + wvs + "_quad_lin"
        polopts["ztitle"] = "$Q_l$"
        store_data(
            vname,
            data={"x": rdtime, "y": reduce_tres(cpowb * mask_final, r), "v": yax},
            attr_dict=polopts,
        )
        wav_data_set_options(vname, polopts)

        # Circular polarization cross-correlation
        wa = wave[..., 0]
        wb = wave[..., 1]
        cpol, cpow, cpowb, _ = cross_corr_wavelet(wa, wb, wid, gaussian=True)

        vname = name + wvs + "_gam_cir"
        ratopts["ztitle"] = "$g_p$"
        store_data(
            vname,
            data={"x": rdtime, "y": reduce_tres(cpol * mask_final, r), "v": yax},
            attr_dict=ratopts,
        )
        wav_data_set_options(vname, ratopts)

        vname = name + wvs + "_coin_cir"
        polopts["ztitle"] = "$C_p$"
        store_data(
            vname,
            data={"x": rdtime, "y": reduce_tres(cpow * mask_final, r), "v": yax},
            attr_dict=polopts,
        )
        wav_data_set_options(vname, polopts)

        vname = name + wvs + "_quad_cir"
        polopts["ztitle"] = "$Q_p$"
        store_data(
            vname,
            data={"x": rdtime, "y": reduce_tres(cpowb * mask_final, r), "v": yax},
            attr_dict=polopts,
        )
        wav_data_set_options(vname, polopts)

    if cross2 and nk >= 3:
        logging.info("Computing cross-correlation cross2")
        # Parallel-perpendicular cross-correlation
        wa = wave[..., 2]  # parallel component
        wb = wave[..., 0] + 1j * wave[..., 1]  # right-handed

        c2pol, c2pow, c2powb, _ = cross_corr_wavelet(wa, wb, wid, gaussian=False)

        vname = name + wvs + "_gam_pr"
        ratopts["ztitle"] = "$g_{pr}$"
        store_data(
            vname,
            data={"x": rdtime, "y": reduce_tres(c2pol * mask_final, r), "v": yax},
            attr_dict=ratopts,
        )
        wav_data_set_options(vname, ratopts)

        vname = name + wvs + "_coin_pr"
        polopts["ztitle"] = "$C_{pr}$"
        store_data(
            vname,
            data={"x": rdtime, "y": reduce_tres(c2pow * mask_final, r), "v": yax},
            attr_dict=polopts,
        )
        wav_data_set_options(vname, polopts)

        vname = name + wvs + "_quad_pr"
        polopts["ztitle"] = "$Q_{pr}$"
        store_data(
            vname,
            data={"x": rdtime, "y": reduce_tres(c2powb * mask_final, r), "v": yax},
            attr_dict=polopts,
        )
        wav_data_set_options(vname, polopts)

        # Left-handed
        wb = wave[..., 0] - 1j * wave[..., 1]
        c2pol, c2pow, c2powb, _ = cross_corr_wavelet(wa, wb, wid, gaussian=False)

        vname = name + wvs + "_gam_pl"
        ratopts["ztitle"] = "$g_{pl}$"
        store_data(
            vname,
            data={"x": rdtime, "y": reduce_tres(c2pol * mask_final, r), "v": yax},
            attr_dict=ratopts,
        )
        wav_data_set_options(vname, ratopts)

        vname = name + wvs + "_coin_pl"
        polopts["ztitle"] = "$C_{pl}$"
        store_data(
            vname,
            data={"x": rdtime, "y": reduce_tres(c2pow * mask_final, r), "v": yax},
            attr_dict=polopts,
        )
        wav_data_set_options(vname, polopts)

        vname = name + wvs + "_quad_pl"
        polopts["ztitle"] = "$Q_{pl}$"
        store_data(
            vname,
            data={"x": rdtime, "y": reduce_tres(c2powb * mask_final, r), "v": yax},
            attr_dict=polopts,
        )
        wav_data_set_options(vname, polopts)

    logging.info(f"Finished wav_data for {name}")

    # Return dictionary with results
    result = {
        "wave": wave,
        "power": wpow,
        "period": period,
        "yax": yax,
        "time": time,
        "returned_rotmats": returned_rotmats,
    }

    return result



from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot

from .config import CONFIG


def load(
    trange=["2004-11-5", "2004-11-6"],
    instrument="lena",
    datatype="k0",
    suffix='',
    prefix='',
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    time_clip=False,
    force_download=False,
):
    """
    Load IMAGE data.

    This function loads data from the Image mission; this function is not meant
    to be called directly; instead, see the wrappers:
    pyspedas.projects.image.lena
    pyspedas.projects.image.mena
    pyspedas.projects.image.hena
    pyspedas.projects.image.rpi
    pyspedas.projects.image.euv
    pyspedas.projects.image.fuv
    pyspedas.projects.image.orbit


    Parameters
    ----------
    trange : list of str, optional
        Time range for the data in the format ['start_date', 'end_date'].
        Defaults to ['2004-11-5', '2004-11-6'].
    instrument : str, optional
        Instrument name. Defaults to 'lena'.
    datatype : str, optional
        Data type. Defaults to 'k0'.
    suffix : str, optional
        Suffix to be added to the variable names. Defaults to ''.
    prefix: str, optional
        Prefix to be added to the variable names. Defaults to ''.
    get_support_data : bool, optional
        Flag indicating whether to retrieve support data. Defaults to False.
    varformat : str, optional
        Variable format. Defaults to None.
    varnames : list of str, optional
        List of variable names to load. Defaults to [].
    downloadonly : bool, optional
        Flag indicating whether to only download the files without loading the data. Defaults to False.
    notplot : bool, optional
        Flag indicating whether to return the loaded data without plotting. Defaults to False.
    no_update : bool, optional
        Flag indicating whether to skip updating the data files. Defaults to False.
    time_clip : bool, optional
        Flag indicating whether to clip the loaded data to the specified time range. Defaults to False.
    force_download: bool
        Download file even if local version is more recent than server version
        Default: False

    Returns
    -------
    list or dict
        If `downloadonly` is True, returns a list of downloaded file names.
        If `notplot` is True, returns a dictionary of loaded data variables.
        Otherwise, returns a list of loaded data variables.
    """

    if instrument == "lena":
        pathformat = (
            instrument
            + "/"
            + instrument
            + "_"
            + datatype
            + "/%Y/im_"
            + datatype
            + "_"
            + instrument
            + "_%Y%m%d_v??.cdf"
        )
    elif instrument == "mena":
        pathformat = (
            instrument
            + "/"
            + instrument
            + "_"
            + datatype
            + "/%Y/im_"
            + datatype
            + "_"
            + instrument
            + "_%Y%m%d_v??.cdf"
        )
    elif instrument == "hena":
        pathformat = (
            instrument
            + "/"
            + instrument
            + "_"
            + datatype
            + "/%Y/im_"
            + datatype
            + "_"
            + instrument
            + "_%Y%m%d_v??.cdf"
        )
    elif instrument == "rpi":
        pathformat = (
            instrument
            + "/"
            + instrument
            + "_"
            + datatype
            + "/%Y/im_"
            + datatype
            + "_"
            + instrument
            + "_%Y%m%d_v??.cdf"
        )
    elif instrument == "euv":
        pathformat = (
            instrument
            + "/"
            + instrument
            + "_"
            + datatype
            + "/%Y/im_"
            + datatype
            + "_"
            + instrument
            + "_%Y%m%d_v??.cdf"
        )
    elif instrument == "fuv":
        pathformat = (
            instrument
            + "/wic_"
            + datatype
            + "/%Y/im_"
            + datatype
            + "_wic_%Y%m%d_v??.cdf"
        )
    elif instrument == "orbit":
        if datatype == "def_or":
            pathformat = instrument + "/def_or/%Y/im_or_def_%Y%m%d_v??.cdf"
        elif datatype == "pre_or":
            pathformat = instrument + "/pre_or/%Y/im_or_pre_%Y%m%d_v??.cdf"

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange)

    out_files = []

    files = download(
        remote_file=remote_names,
        remote_path=CONFIG["remote_data_dir"],
        local_path=CONFIG["local_data_dir"],
        no_download=no_update,
        force_download=force_download,
    )
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(
        out_files,
        suffix=suffix,
        prefix=prefix,
        get_support_data=get_support_data,
        varformat=varformat,
        varnames=varnames,
        notplot=notplot,
    )

    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix="")

    return tvars
